
#include <iostream>
#include <ctime>
#include <random>

#include "catch.hpp"

#include "MassSpectrum.hpp"
#include "MassSpectrumList.hpp"
#include "globals/globals.hpp"

#include <QDebug>
#include <QFile>

using std::cout;

#define TEST_DATA_DIR "/home/rusconi/devel/msxpertsuite/development/tests/data"

//qDebug() << __FILE__ << __LINE__ << __FUNCTION__ 
//<< "Should compile libmass tests.";

namespace msXpSlibmass
{

	TEST_CASE("mass spectra can have a title ", "[MassSpectrum]")
	{
		QString title("spectrum title");
		MassSpectrum spectrum(title);

		REQUIRE(spectrum.title() == "spectrum title");
	}


	TEST_CASE("mass spectra have rt and dt", "[MassSpectrum]")
	{
		MassSpectrum spectrum;

		double rt = 4.258;
		double dt = 12.9854;

		spectrum.rrt() = rt;
		spectrum.rdt() = dt;

		REQUIRE(spectrum.rt() == rt);
		REQUIRE(spectrum.dt() == dt);
	}


	TEST_CASE("mass spectra have decimal places", "[MassSpectrum]")
	{
		MassSpectrum spectrum;
		int decimalPlaces = 5;

		spectrum.setDecimalPlaces(decimalPlaces);
		REQUIRE(spectrum.decimalPlaces() == decimalPlaces);

		decimalPlaces = 10;

		spectrum.rdecimalPlaces() = decimalPlaces;
		REQUIRE(spectrum.decimalPlaces() == decimalPlaces);
	}


	TEST_CASE("Mass spectra can be initialized with double lists", "[MassSpectrum]") 
	{
		MassSpectrum massSpectrum;

		REQUIRE(massSpectrum.size() == 0);

		SECTION("The spectrum is initialized with mz and i double lists")
		{
			QList<double> mzList={1.00, 2.100, 3.200, 4.300, 5.400};
			QList<double> iList={1300, 2600, 5200, 10400, 20800};
			Q_ASSERT(mzList.size() == iList.size());
			int itemCount = mzList.size();

			massSpectrum.initialize(mzList, iList);

			REQUIRE(massSpectrum.size() == itemCount);

			int failedMzComparisons = 0;
			int failedIComparisons = 0;

			for(int iter = 0; iter < itemCount; ++iter)
			{
				MassPeak *curMassPeak = massSpectrum.at(iter);

				if(curMassPeak->mz() != mzList.at(iter))
					++failedMzComparisons;

				if(curMassPeak->i() != iList.at(iter))
					++failedIComparisons;
			}

			REQUIRE(failedMzComparisons == 0);
			REQUIRE(failedIComparisons == 0);
		}

		SECTION("The spectrum is initialized with mz and i double lists "
				"and start-end range values")
		{
			QList<double> mzList={725.639, 987.4569, 1235.1234, 1236.1234, 2598.456, 2599.457, 3569.12458};
			QList<double> iList={1250, 2500, 5000, 10000, 20000, 40000, 80000};
			Q_ASSERT(mzList.size() == iList.size());
			int itemCount = mzList.size();

			double start = 1235.12341;
			double end = 2599.456;

			massSpectrum.initialize(mzList, iList, start, end);

			REQUIRE(massSpectrum.size() == 2);

			REQUIRE(massSpectrum[0]->mz() == 1236.1234);
			REQUIRE(massSpectrum[0]->i() == 10000);

			REQUIRE(massSpectrum[1]->mz() == 2598.456);
			REQUIRE(massSpectrum[1]->i() == 20000);
		}
	}


	TEST_CASE("Mass spectra can be initialized with double "
			"lists and a drift time value", "[MassSpectrum]") 
	{
		MassSpectrum massSpectrum;

		REQUIRE(massSpectrum.size() == 0);

		SECTION("The spectrum is initialized with mz and i double "
				"lists and one drift time value")
		{
			QList<double> mzList={1.00, 2.100, 3.200, 4.300, 5.400};
			QList<double> iList={1300, 2600, 5200, 10400, 20800};
			Q_ASSERT(mzList.size() == iList.size());
			int itemCount = mzList.size();

			double driftTime = 27.7895;

			massSpectrum.initialize(mzList, iList, driftTime);

			REQUIRE(massSpectrum.size() == itemCount);
			REQUIRE(massSpectrum.dt() == driftTime);

			int failedMzComparisons = 0;
			int failedIComparisons = 0;

			for(int iter = 0; iter < itemCount; ++iter)
			{
				MassPeak *curMassPeak = massSpectrum.at(iter);

				if(curMassPeak->mz() != mzList.at(iter))
					++failedMzComparisons;

				if(curMassPeak->i() != iList.at(iter))
					++failedIComparisons;
			}

			REQUIRE(failedMzComparisons == 0);
			REQUIRE(failedIComparisons == 0);
		}

		SECTION("The spectrum is initialized with mz and i double lists, "
				"drift time value and and start-end range values")
		{
			QList<double> mzList={725.639, 987.4569, 1235.1234, 1236.1234, 2598.456, 2599.457, 3569.12458};
			QList<double> iList={1250, 2500, 5000, 10000, 20000, 40000, 80000};
			Q_ASSERT(mzList.size() == iList.size());
			int itemCount = mzList.size();

			double driftTime = 27.7895;

			double start = 1235.12341;
			double end = 2599.456;

			massSpectrum.initialize(mzList, iList, driftTime, start, end);

			REQUIRE(massSpectrum.size() == 2);
			REQUIRE(massSpectrum.dt() == driftTime);

			REQUIRE(massSpectrum[0]->mz() == 1236.1234);
			REQUIRE(massSpectrum[0]->i() == 10000);

			REQUIRE(massSpectrum[1]->mz() == 2598.456);
			REQUIRE(massSpectrum[1]->i() == 20000);
		}
	}


	TEST_CASE("Mass spectra can be initialized with double vectors "
			"and start-end range values", "[MassSpectrum]") 
	{
		MassSpectrum massSpectrum;

		REQUIRE(massSpectrum.size() == 0);

		SECTION("The spectrum is initialized with mz and i double "
				"vectors and no start-end range values")
		{
			QVector<double> mzVector={1.00, 2.100, 3.200, 4.300, 5.400};
			QVector<double> iVector={1300, 2600, 5200, 10400, 20800};
			Q_ASSERT(mzVector.size() == iVector.size());
			int itemCount = mzVector.size();

			massSpectrum.initialize(mzVector, iVector);

			REQUIRE(massSpectrum.size() == itemCount);

			int failedMzComparisons = 0;
			int failedIComparisons = 0;

			for(int iter = 0; iter < itemCount; ++iter)
			{
				MassPeak *curMassPeak = massSpectrum.at(iter);

				if(curMassPeak->mz() != mzVector.at(iter))
					++failedMzComparisons;

				if(curMassPeak->i() != iVector.at(iter))
					++failedIComparisons;
			}

			REQUIRE(failedMzComparisons == 0);
			REQUIRE(failedIComparisons == 0);
		}


		SECTION("The spectrum is initialized with mz and i double "
				"vectors and start-end range values")
		{
			QVector<double> mzVector={725.639, 987.4569, 1235.1234, 1236.1234, 2598.456, 2599.457, 3569.12458};
			QVector<double> iVector={1250, 2500, 5000, 10000, 20000, 40000, 80000};
			Q_ASSERT(mzVector.size() == iVector.size());
			int itemCount = mzVector.size();

			double start = 1235.12341;
			double end = 2599.456;

			massSpectrum.initialize(mzVector, iVector, start, end);

			REQUIRE(massSpectrum.size() == 2);

			REQUIRE(massSpectrum[0]->mz() == 1236.1234);
			REQUIRE(massSpectrum[0]->i() == 10000);

			REQUIRE(massSpectrum[1]->mz() == 2598.456);
			REQUIRE(massSpectrum[1]->i() == 20000);
		}
	}


	TEST_CASE("Mass spectra can be initialized with simple XY-formatted text.", "[MassSpectrum]") 
	{
		MassSpectrum massSpectrum;

		REQUIRE(massSpectrum.size() == 0);

		QStringList xyFormatStringList;
		xyFormatStringList	<< "199.928558349609 1840.123456700000"
			<< "199.934738159180 2471.456780000000"
			<< "199.940917968750 3497.000000056985"
			<< "199.947082519531 5236.000000000000"
			<< "199.953262329102 8631.000000000000"
			<< "199.959442138672 31505.000000000000"
			<< "199.965606689453 75716.000000000000"
			<< "199.971786499023 79882.000000000000"
			<< "199.977951049805 47849.000000000000"
			<< "199.984130859375 33539.000000000000"
			<< "199.990295410156 29096.000000000000"
			<< "199.996475219727 29104.000000000000"
			<< "200.002655029297 26650.000000000000"
			<< "200.008819580078 24243.000000000000"
			<< "200.014999389648 25983.000000000000"
			<< "200.021179199219 28867.000000000000"
			<< "200.027359008789 23874.000000000000"
			<< "200.033523559570 19043.000000000000"
			<< "200.039703369141 16450.000000000000"
			<< "200.045883178711 17226.000000000000"
			<< "200.052062988281 30648.000000000000"
			<< "200.058242797852 44396.000000000000"
			<< "200.064407348633 34241.000000000000"
			<< "200.070587158203 23241.000000000000"
			<< "200.076766967773 17745.000000000000"
			<< "200.082946777344 16258.000000000000"
			<< "200.089126586914 30343.000000000000";

		int xyFormatStringListSize = xyFormatStringList.size();

		QString xyFormatString = xyFormatStringList.join("\n");

		massSpectrum.initialize(xyFormatString);

		REQUIRE(xyFormatStringListSize == massSpectrum.size());

		int failedMzComparisons = 0;
		int failedIComparisons = 0;

		for(int iter = 0; iter < xyFormatStringListSize; ++iter)
		{
			MassPeak *iterMassPeak = massSpectrum.at(iter);
			MassPeak *massPeak = new MassPeak(xyFormatStringList.at(iter));

			if(iterMassPeak->mz() != massPeak->mz())
				++failedMzComparisons;

			if(iterMassPeak->i() != massPeak->i())
				++failedIComparisons;
		}

		REQUIRE(failedMzComparisons == 0);
		REQUIRE(failedIComparisons == 0);
	}


	TEST_CASE("All the mass peaks of a mass spectrum can be deleted.", "[MassSpectrum]")
	{
		MassSpectrum massSpectrum;
		REQUIRE(massSpectrum.size() == 0);

		QVector<double> mzVector={725.639, 987.4569, 1235.1234, 1236.1234, 2598.456, 2599.457, 3569.12458};
		QVector<double> iVector={1250, 2500, 5000, 10000, 20000, 40000, 80000};
		Q_ASSERT(mzVector.size() == iVector.size());
		int itemCount = mzVector.size();

		massSpectrum.initialize(mzVector, iVector);
		REQUIRE(massSpectrum.size() == itemCount);

		massSpectrum.deleteMassPeaks();
		REQUIRE(massSpectrum.size() == 0);
	}


	TEST_CASE("Mass spectra can be reset to an empty state.", "[MassSpectrum]")
	{
		MassSpectrum massSpectrum;
		REQUIRE(massSpectrum.size() == 0);

		QVector<double> mzVector={725.639, 987.4569, 1235.1234, 1236.1234, 2598.456, 2599.457, 3569.12458};
		QVector<double> iVector={1250, 2500, 5000, 10000, 20000, 40000, 80000};
		Q_ASSERT(mzVector.size() == iVector.size());
		int itemCount = mzVector.size();

		massSpectrum.initialize(mzVector, iVector);
		REQUIRE(massSpectrum.size() == itemCount);

		massSpectrum.reset();
		REQUIRE(massSpectrum.size() == 0);
	}

	TEST_CASE("When combining mass spectra, binning might be required. "
			"Testing that feature (non-binnable test).", "[MassSpectrum]")
	{
		MassSpectrum massSpectrum;
		REQUIRE(massSpectrum.size() == 0);

		double seedMzValue = 694.5166130065;
		QVector<double> mzVector;
		mzVector << seedMzValue;
		double lastMzValue = seedMzValue;
		double newMzValue;

		// Craft on purpose non-binnable mz data.

		for(int iter = 0; iter < 24; ++iter)
		{
			if(iter % 2 == 0)
			{
				newMzValue = lastMzValue + MassSpectrum::m_maxAvgMzBinSize + 0.01;
				mzVector << newMzValue;
				lastMzValue = newMzValue;
			}
			else
			{
				newMzValue = lastMzValue + MassSpectrum::m_maxAvgMzBinSize;
				mzVector << newMzValue;
				lastMzValue = newMzValue;
			}
		}

		QVector<double> iVector={
			204.0000000000, 277.0000000000, 300.0000000000, 286.0000000000, 273.0000000000, 276.0000000000, 268.0000000000, 235.0000000000, 223.0000000000, 195.0000000000, 220.0000000000, 233.0000000000, 249.0000000000, 247.0000000000, 263.0000000000, 299.0000000000, 335.0000000000, 354.0000000000, 317.0000000000, 249.0000000000, 217.0000000000, 217.0000000000, 227.0000000000, 231.0000000000, 249.0000000000};

		if(mzVector.size() != iVector.size())
			qFatal("Fatal error at %s@%d -- %s. "
					"The mz and i vectors have not the same size."
					"Program aborted.",
					__FILE__, __LINE__, __FUNCTION__);

		int itemCount = mzVector.size();

		massSpectrum.initialize(mzVector, iVector);
		REQUIRE(massSpectrum.size() == itemCount);

		QList<double> binList;
		massSpectrum.fillInBins(&binList);

		REQUIRE(massSpectrum.isBinnable(binList) == false);
	}


	TEST_CASE("When combining mass spectra, binning might be required. "
			"Testing that feature (binnable test).", "[MassSpectrum]")
	{
		// Prepare the mass data to initialize the mass spectrum.
		// Prepare a local test bin list that must be discovered to be identical
		// to the bin list computed by MassSpectrum::fillInBins()

		QList<double> mzList = 
		{
			694.5166264003, 694.5333824250, 694.5501386518, 694.5668950807, 694.5836517118, 694.6004085450, 694.6171655803, 694.6339228177, 694.6506802573, 694.6674378989, 694.6841957428, 694.7009537887, 694.7177120368, 694.7344704869, 694.7512291393, 694.7679879937, 694.7847470503, 694.8015063090, 694.8182657698, 694.8350254327};

		QList<double> iList = 
		{
			133.0000000000, 195.0000000000, 225.0000000000, 220.0000000000, 286.0000000000, 332.0000000000, 282.0000000000, 164.0000000000, 140.0000000000, 156.0000000000, 153.0000000000, 89.0000000000, 104.0000000000, 143.0000000000, 196.0000000000, 295.0000000000, 396.0000000000, 399.0000000000, 369.0000000000, 345.0000000000};

		if(mzList.size() != iList.size())
			qFatal("Fatal error at %s@%d -- %s. "
					"The mz and i lists have not the same size."
					"Program aborted.",
					__FILE__, __LINE__, __FUNCTION__);

		int itemCount = mzList.size();

		MassSpectrum massSpectrum;
		massSpectrum.initialize(mzList, iList);

		// Check that initialization is correct.
		REQUIRE(massSpectrum.size() == itemCount);

		// Prepare the bin list and ask the spectrum to fill it.
		QList<double> binList;
		massSpectrum.fillInBins(&binList);

		int binListCount = binList.size();

		// The bin list count must be one item less than the size of the mass
		// spectrum.
		REQUIRE(binListCount == itemCount - 1);

		// Prepare a local copy of the bin list.
		QList<double> testBinList;

		for(int iter = 1; iter < itemCount; ++iter)
		{
			double binValue = mzList.at(iter) - mzList.at(iter - 1) ;
			testBinList << binValue;
		}

		SECTION("Test the filling-in of the bins", "[MassSpectrum]")
		{
			// The bin lists must be identical. Factorize all the comparisons into
			// a single value. If all the values are indeed identical, then
			// failedComparisons should stay = 0.

			// Both bin lists must have the same item number.
			REQUIRE(binListCount == testBinList.size());


			int failedComparisons = 0;
			for(int iter = 0; iter < binListCount; ++iter)
			{
				if(binList.at(iter) != testBinList.at(iter))
					++failedComparisons;
			}

			// Finally test if the bin lists were identical.
			REQUIRE(failedComparisons == 0);
		}

		SECTION("Test the statistics of the bins (avg and stdDev).", "[MassSpectrum]")
		{

			// Compute the test bin average value.

			double testBinListSum = 0;

			for(int iter = 0; iter < testBinList.size(); ++iter)
				testBinListSum += testBinList.at(iter);

			double testBinSizeAvg = testBinListSum / testBinList.size();

			// Now perform the actual test.
			REQUIRE(testBinSizeAvg == massSpectrum.binSizeAvg(binList));

			// Now to the standard deviation. Compute the value for the test bin list.

			double testBinSizeVarN = 0;

			for(int iter = 0; iter < testBinList.size(); ++iter)
				testBinSizeVarN += (testBinList.at(iter) - testBinSizeAvg) * 
					(testBinList.at(iter) - testBinSizeAvg);

			double testBinSizeVar = testBinSizeVarN / testBinList.size();

			double testBinStdDev = sqrt(testBinSizeVar);

			// Finally compare the test value with the mass spectrum-generated value.
			REQUIRE(testBinStdDev == massSpectrum.binSizeStdDev(binList));

			// And now finish up with the binnability test.

			bool testBinnability = (testBinSizeAvg <= MassSpectrum::m_maxAvgMzBinSize) &&
				(testBinStdDev <= MassSpectrum::m_maxBinSizeStdDev);

			REQUIRE(testBinnability == massSpectrum.isBinnable(binList));
		}

		SECTION("Test the setting-up of binnability of spectra.", "[MassSpectrum]")
		{

			// We know that the data used to seed the new spectra below are correct
			// for binnability.

			MassSpectrum binMassSpectrum;

			MassSpectrumList massSpectra;

			// We need two allocated mass spectrum instances in the mass spectrum
			// list.

			MassSpectrum *newSpectrum = new MassSpectrum;
			newSpectrum->initialize(mzList, iList);
			massSpectra.append(newSpectrum);

			newSpectrum = new MassSpectrum;
			newSpectrum->initialize(mzList, iList);
			massSpectra.append(newSpectrum);

			bool res = binMassSpectrum.setupBinnability(massSpectra);
			REQUIRE(res == true);

			// At this point, binMassSpectrum must have as many MassPeak elements as
			// what was in newSpectrum, contrary to the binList which has (itemCount -
			// 1) bins. This is because when populating binMassSpectrum in the
			// setupBinnability() function, the smallest mz value found in all the
			// mass spectra of the MassSpectrumList is first seeded as such, and is
			// thus the first mass peak item of the mass spectrum. Then, for each bin,
			// a new mass peak is appended to the spectrum. At the end, there are bins
			// + 1 mass peaks in the spectrum, which is identical to itemCount.

			REQUIRE(binMassSpectrum.size() == itemCount);

			delete massSpectra.at(1);
			delete massSpectra.at(0);

			massSpectra.clear();
		}

		SECTION("Test the binIndex feature: finding the bin in "
				"which to account a new m/z value.", "[MassSpectrum]")
		{
			int mzIndex = 0;
			double newMzValue = 0;

			// Before the first mz item of the list.
			newMzValue = mzList.at(mzIndex) - 0.0000000001;
			REQUIRE(massSpectrum.binIndex(newMzValue) == mzIndex);

			// In between the borders of the list
			mzIndex = 10;
			double binSize = mzList.at(mzIndex + 1) - mzList.at(mzIndex);

			// Right on, the index should stay the same.
			newMzValue = mzList.at(mzIndex);
			REQUIRE(massSpectrum.binIndex(newMzValue) == mzIndex);

			// Remove to the mz value right enough weight to fall in previous bin.
			newMzValue = mzList.at(mzIndex) - 0.0000000001;
			REQUIRE(massSpectrum.binIndex(newMzValue) == mzIndex - 1);

			// Add to the mz value not enough weight to fall in next bin.
			newMzValue = mzList.at(mzIndex) + binSize - 0.0000000001;
			REQUIRE(massSpectrum.binIndex(newMzValue) == mzIndex);

			newMzValue = mzList.at(mzIndex) + binSize;
			REQUIRE(massSpectrum.binIndex(newMzValue) == mzIndex + 1);

			// After the last mz item of the list.
			newMzValue = mzList.at(itemCount - 1) + 0.0000000001;
			REQUIRE(massSpectrum.binIndex(newMzValue) == massSpectrum.size() - 1);
		}
	}


	TEST_CASE("When combining mass spectra, find an mz shift between "
			"spectra might be required. Testing that feature.", "[MassSpectrum]")
	{
		// Prepare the mass data to initialize the mass spectrum.
		// Prepare a local test bin list that must be discovered to be identical
		// to the bin list computed by MassSpectrum::fillInBins()

		QList<double> mzList = 
		{
			694.5166264003,
			694.5333824250,
			694.5501386518,
			694.5668950807,
			694.5836517118,
			694.6004085450,
			694.6171655803,
			694.6339228177,
			694.6506802573,
			694.6674378989,
			694.6841957428,
			694.7009537887,
			694.7177120368,
			694.7344704869,
			694.7512291393,
			694.7679879937,
			694.7847470503,
			694.8015063090,
			694.8182657698,
			694.8350254327};

		QList<double> iList = 
		{
			133.0000000000,
			195.0000000000,
			225.0000000000,
			220.0000000000,
			286.0000000000,
			332.0000000000,
			282.0000000000,
			164.0000000000,
			140.0000000000,
			156.0000000000,
			153.0000000000,
			89.0000000000,
			104.0000000000,
			143.0000000000,
			196.0000000000,
			295.0000000000,
			396.0000000000,
			399.0000000000,
			369.0000000000,
			345.0000000000};

		if(mzList.size() != iList.size())
			qFatal("Fatal error at %s@%d -- %s. "
					"The mz and i lists have not the same size."
					"Program aborted.",
					__FILE__, __LINE__, __FUNCTION__);

		int itemCount = mzList.size();

		MassSpectrum massSpectrum;
		massSpectrum.initialize(mzList, iList);

		// Check that initialization is correct.
		REQUIRE(massSpectrum.size() == itemCount);

		// Now prepare the other mass spectrum that is to be combined into the
		// previous. We just increase the first mass spectrum mz values by a fake
		// shift value.

		double fakeShift = 0.001234;
		QList<double> mzListOther;

		for(int iter = 0; iter < itemCount; ++iter)
			mzListOther << mzList.at(iter) + fakeShift;

		MassSpectrum massSpectrumOther;
		massSpectrumOther.initialize(mzListOther, iList);

		// Check that initialization is correct.
		REQUIRE(massSpectrumOther.size() == itemCount);
		REQUIRE(massSpectrumOther.at(0)->mz() - fakeShift == massSpectrum.at(0)->mz());

		// Now test the shift calculation.
		double result = 0;

		result = massSpectrum.determineMzShift(massSpectrumOther);
		// It is odd that REQUIRE(result == fakeShift) fails!
		REQUIRE(massSpectrumOther.at(0)->mz() - result == massSpectrum.at(0)->mz());

		result = massSpectrum.determineMzShift(massSpectrumOther.at(0)->mz());
		// It is odd that REQUIRE(result == fakeShift) fails!
		REQUIRE(massSpectrumOther.at(0)->mz() - result == massSpectrum.at(0)->mz());
	}


	TEST_CASE("TIC calculations.", "[MassSpectrum]")
	{
		QList<double> mzList = 
		{
			694.5166264003, 694.5333824250, 694.5501386518, 694.5668950807, 694.5836517118, 694.6004085450, 694.6171655803, 694.6339228177, 694.6506802573, 694.6674378989, 694.6841957428, 694.7009537887, 694.7177120368, 694.7344704869, 694.7512291393, 694.7679879937, 694.7847470503, 694.8015063090, 694.8182657698, 694.8350254327};

		QList<double> iList = 
		{
			133.0000000000, 195.0000000000, 225.0000000000, 220.0000000000, 286.0000000000, 332.0000000000, 282.0000000000, 164.0000000000, 140.0000000000, 156.0000000000, 153.0000000000, 89.0000000000, 104.0000000000, 143.0000000000, 196.0000000000, 295.0000000000, 396.0000000000, 399.0000000000, 369.0000000000, 345.0000000000};

		if(mzList.size() != iList.size())
			qFatal("Fatal error at %s@%d -- %s. "
					"The mz and i lists have not the same size."
					"Program aborted.",
					__FILE__, __LINE__, __FUNCTION__);

		int itemCount = mzList.size();

		MassSpectrum massSpectrum;
		massSpectrum.initialize(mzList, iList);

		// Check that initialization is correct.
		REQUIRE(massSpectrum.size() == itemCount);

		SECTION("Compute the TIC with no mz borders.", "[MassSpectrum]")
		{
			// We computed the TIC in SpeedCrunch: (4622.00000000).
			REQUIRE(massSpectrum.tic() == 4622.00000000);
		}

		SECTION("Compute the TIC with mz range values.", "[MassSpectrum]")
		{
			// Want to leave out the first four and last four mz values of the
			// spectrum. SpeedCrunch value: 2340.00000000
			double mzStart = 694.5668950808; // slightly more than fourth from top
			double mzEnd = 694.7847470502; // slightly less fourth from bottom

			REQUIRE(massSpectrum.tic(mzStart, mzEnd) == 2340.00000000);
		}
	}


	TEST_CASE("Mass spectrum combination tests.", "[MassSpectrum]")
	{
		QList<double> mzList = 
		{
			694.5166264003, 694.5333824250, 694.5501386518, 694.5668950807, 694.5836517118, 694.6004085450, 694.6171655803, 694.6339228177, 694.6506802573, 694.6674378989, 694.6841957428, 694.7009537887, 694.7177120368, 694.7344704869, 694.7512291393, 694.7679879937, 694.7847470503, 694.8015063090, 694.8182657698, 694.8350254327};

		QList<double> iList = 
		{
			133.0000000000, 195.0000000000, 225.0000000000, 220.0000000000, 286.0000000000, 332.0000000000, 282.0000000000, 164.0000000000, 140.0000000000, 156.0000000000, 153.0000000000, 89.0000000000, 104.0000000000, 143.0000000000, 196.0000000000, 295.0000000000, 396.0000000000, 399.0000000000, 369.0000000000, 345.0000000000};

		if(mzList.size() != iList.size())
			qFatal("Fatal error at %s@%d -- %s. "
					"The mz and i lists have not the same size."
					"Program aborted.",
					__FILE__, __LINE__, __FUNCTION__);

		int itemCount = mzList.size();

		MassSpectrum massSpectrum;
		massSpectrum.initialize(mzList, iList);

		// Create an identical copy to later check the differences.
		MassSpectrum otherMassSpectrum;
		otherMassSpectrum.initialize(mzList, iList);

		// Check that initialization is correct.
		REQUIRE(massSpectrum.size() == itemCount);

		SECTION("Combine a mass peak into the mass spectrum when "
				"mz is already in the spectrum.", "[MassSpectrum]")
		{

			// In this test, we should not create a new MassPeak in the spectrum, but
			// increment the i value of one of the MassPeak items.
			std::default_random_engine generator;
			std::uniform_int_distribution<int>distribution(10, itemCount -10);
			int mzIndex = distribution(generator);  // generates number in the range 1..6srand(time(NULL));
			//cout << "Random index:" << mzIndex << "\n";
			int prevI = massSpectrum.at(mzIndex)->i();
			//cout << std::setprecision(15) << "m/z value for index:" << massSpectrum.at(mzIndex)->mz() << "\n";
			//cout << std::setprecision(15) << "Intensity for index:" << prevI << "\n";

			double newMzValue = massSpectrum.at(mzIndex)->mz();

			MassPeak massPeak(newMzValue, 1234);

			massSpectrum.combine(massPeak);

			// The only differing MassPeak item should be at index mzIndex. Let's
			// check that.

			QMap<int, double> changedIdxVsMzMap;
			QMap<int, double> changedIdxVsIMap;

			for(int iter = 0; iter < itemCount; ++iter)
			{
				double mz = massSpectrum.at(iter)->mz();
				double i = massSpectrum.at(iter)->i();

				double otherMz = otherMassSpectrum.at(iter)->mz();
				double otherI = otherMassSpectrum.at(iter)->i();

				if(mz != otherMz)
					changedIdxVsMzMap.insert(iter, mz);

				if(i != otherI)
					changedIdxVsIMap.insert(iter, i);
			}

			REQUIRE(changedIdxVsMzMap.size() == 0);
			REQUIRE(changedIdxVsIMap.size() == 1);

			REQUIRE(changedIdxVsIMap.first() /* value in the map */ == prevI + 1234);
			REQUIRE(changedIdxVsIMap.key(changedIdxVsIMap.first()) == mzIndex);

			// At this point we have tested that we have updated the i value of a
			// MassPeak located at the intended position.
		}

		SECTION("Combine a mass peak into the mass spectrum when "
				"mz is not already in the spectrum.", "[MassSpectrum]")
		{
			// In this test, we know we should create a new MassPeak item inside the
			// spectrum one index after the targeted one.

			std::default_random_engine generator;
			std::uniform_int_distribution<int>distribution(10, itemCount - 10);
			int mzIndex = distribution(generator);  // generates number in the range 1..6srand(time(NULL));
			//cout << "Random index:" << mzIndex << "\n";
			double origPrevMz = massSpectrum.at(mzIndex)->mz();
			double origNextMz = massSpectrum.at(mzIndex + 1)->mz();

			double origPrevI = massSpectrum.at(mzIndex)->i();
			double origNextI = massSpectrum.at(mzIndex + 1)->i();

			//cout << std::setprecision(15) << "orig prev: (" << origPrevMz << "," << origPrevI  << ")\n";
			//cout << std::setprecision(15) << "orig next: (" << origNextMz << "," << origNextI  << ")\n";

			// We want to insert the new MassPeak between origPrevMz and origNextMz.
			double newMzValue = origPrevMz + (origNextMz - origPrevMz) - 0.0000000001;

			MassPeak massPeak(newMzValue, 1234);

			massSpectrum.combine(massPeak);

			// Now that spectrum should be longer by one item.

			REQUIRE(massSpectrum.size() == itemCount + 1);

			// The only differing MassPeak item should be at index (mzIndex + 1).
			// Let's check that.

			int failedComparisons = 0;

			for(int iter = 0; iter <= mzIndex; ++iter)
			{
				if(massSpectrum.at(iter)->mz() !=
						otherMassSpectrum.at(iter)->mz())
				{
					//cout << std::setprecision(5) << "failed mz comparison at iter:" << iter << "\n";

					++failedComparisons;
				}

				// Also compare the intensity, since they should not have changed
				// neither.

				if(massSpectrum.at(iter)->i() !=
						otherMassSpectrum.at(iter)->i())
				{
					//cout << std::setprecision(5) << "failed i comparison at iter:" << iter << "\n";

					++failedComparisons;
				}
			}

			REQUIRE(failedComparisons == 0);

			// The inserted MassPeak should be at (mzIndex + 1)
			REQUIRE(massSpectrum.at(mzIndex + 1)->mz() == newMzValue);

			// And now all the remaining items should be identical, but dephased.

			failedComparisons = 0;

			for(int iter = mzIndex + 1; iter < itemCount; ++iter)
			{
				if(otherMassSpectrum.at(iter)->mz() != 
						massSpectrum.at(iter + 1)->mz())
					++failedComparisons;

				if(otherMassSpectrum.at(iter)->i() != 
						massSpectrum.at(iter + 1)->i())
					++failedComparisons;
			}

			REQUIRE(failedComparisons == 0);

			// At this point we have tested the we inserted a new MassPeak at the
			// intended position.
		}

		SECTION("Combine a mass peak into the mass spectrum when "
				"mz is not already in the spectrum and new one should be on top.", "[MassSpectrum]")
		{

			// In this test, we should create a new MassPeak in the spectrum, that
			// should become the first MassPeak of the spectrum after insertion
			// at index 0.

			int mzIndex = 0;
			double origFirstMz = massSpectrum.at(mzIndex)->mz();
			double origFirstI = massSpectrum.at(mzIndex)->i();

			//cout << std::setprecision(15) << "m/z value for index:" << origFirstMz << "\n";
			//cout << std::setprecision(15) << "Intensity for index:" << origFirstI << "\n";


			// Prepare a new mz value that is very slightly inferior to the mz value
			// of the first MassPeak of the spectrum;
			double newMzValue = massSpectrum.at(mzIndex)->mz() - 0.0000000001;
			MassPeak massPeak(newMzValue, 1234);

			massSpectrum.combine(massPeak);

			// The only differing MassPeak item should be at index mzIndex. Let's
			// check that.

			REQUIRE(massSpectrum.at(0)->mz() == newMzValue);
			REQUIRE(massSpectrum.at(0)->i() == 1234);

			// Now, all the other items, should be identical. Let's check that. Skip
			// first mass peak of massSpectrum, because we know it's the new one.
			// Watch the dephasing in the for loop.

			int failedComparisons = 0;

			for(int iter = 1; iter <= itemCount; ++iter)
			{
				double mz = massSpectrum.at(iter)->mz();
				double i = massSpectrum.at(iter)->i();

				double otherMz = otherMassSpectrum.at(iter - 1)->mz();
				double otherI = otherMassSpectrum.at(iter - 1)->i();

				if(mz != otherMz || i != otherI)
					++failedComparisons;
			}

			REQUIRE(failedComparisons == 0);

			// At this point we have tested that we have inserted at index 0 a new
			// MassPeak instance.
		}


		SECTION("Combine a mass peak into the mass spectrum when "
				"mz is not already in the spectrum and new one should be on bottom.", "[MassSpectrum]")
		{

			// In this test, we should create a new MassPeak in the spectrum, that
			// should become the last MassPeak of the spectrum after insertion
			// at index (itemCount - 1).

			int mzIndex = itemCount - 1;
			double origLastMz = massSpectrum.at(mzIndex)->mz();
			double origLastI = massSpectrum.at(mzIndex)->i();

			//cout << std::setprecision(15) << "m/z value for index:" << origLastMz << "\n";
			//cout << std::setprecision(15) << "Intensity for index:" << origLastI << "\n";

			// Prepare a new mz value that is very slightly greater to the mz value
			// of the last MassPeak of the spectrum;
			double newMzValue = massSpectrum.at(mzIndex)->mz() + 0.0000000001;
			MassPeak massPeak(newMzValue, 1234);

			massSpectrum.combine(massPeak);

			// We should have added a new MassPeak instance, the mass spectrum should have grown by
			// 1 unit. Thus, the last index is now itemCount.

			int newItemCount = massSpectrum.size();
			REQUIRE(newItemCount == itemCount + 1);

			REQUIRE(massSpectrum.at(newItemCount - 1)->mz() == newMzValue);
			REQUIRE(massSpectrum.at(newItemCount - 1)->i() == 1234);

			// Now, all the other items, should be identical. Let's check that. Skip
			// last mass peak of massSpectrum, because we know it's the new one.

			int failedComparisons = 0;

			for(int iter = 0; iter < newItemCount - 1 /* skip last item */; ++iter)
			{
				double mz = massSpectrum.at(iter)->mz();
				double i = massSpectrum.at(iter)->i();

				double otherMz = otherMassSpectrum.at(iter)->mz();
				double otherI = otherMassSpectrum.at(iter)->i();

				if(mz != otherMz || i != otherI)
					++failedComparisons;
			}

			REQUIRE(failedComparisons == 0);

			// At this point we have tested that we have inserted a new mass peak as
			// the very last item in the masspectrum.
		}
	}


	TEST_CASE("Mass spectrum binned combination tests.", "[MassSpectrum]")
	{

		QList<double> mzList = 
		{
			694.5166264003, 694.5333824250, 694.5501386518, 694.5668950807, 694.5836517118, 694.6004085450, 694.6171655803, 694.6339228177, 694.6506802573, 694.6674378989, 694.6841957428, 694.7009537887, 694.7177120368, 694.7344704869, 694.7512291393, 694.7679879937, 694.7847470503, 694.8015063090, 694.8182657698, 694.8350254327};

		QList<double> iList = 
		{
			133.0000000000, 195.0000000000, 225.0000000000, 220.0000000000, 286.0000000000, 332.0000000000, 282.0000000000, 164.0000000000, 140.0000000000, 156.0000000000, 153.0000000000, 89.0000000000, 104.0000000000, 143.0000000000, 196.0000000000, 295.0000000000, 396.0000000000, 399.0000000000, 369.0000000000, 345.0000000000};

		if(mzList.size() != iList.size())
			qFatal("Fatal error at %s@%d -- %s. "
					"The mz and i lists have not the same size."
					"Program aborted.",
					__FILE__, __LINE__, __FUNCTION__);

		int itemCount = mzList.size();

		MassSpectrum massSpectrum;
		massSpectrum.initialize(mzList, iList);

		// Create an identical copy to later check the differences.
		MassSpectrum otherMassSpectrum;
		otherMassSpectrum.initialize(mzList, iList);

		// Check that initialization is correct.
		REQUIRE(massSpectrum.size() == itemCount);

		// We know that the data used to seed the new spectra below are correct
		// for binnability.

		// This spectrum will be the host of the bins once computed by looking
		// into the new mass spectra allocated below. Once the bins will be setup
		// in this spectrum, it will be the spectrum into which we test the binned
		// combination of a MassPeak.

		MassSpectrum binMassSpectrum;

		MassSpectrumList massSpectra;

		// We need two allocated mass spectrum instances in the mass spectrum
		// list.

		MassSpectrum *newSpectrum = new MassSpectrum;
		newSpectrum->initialize(mzList, iList);
		massSpectra.append(newSpectrum);

		newSpectrum = new MassSpectrum;
		newSpectrum->initialize(mzList, iList);
		massSpectra.append(newSpectrum);

		bool res = binMassSpectrum.setupBinnability(massSpectra);
		REQUIRE(res == true);

		// At this point, binMassSpectrum must have as many MassPeak elements as
		// what was in newSpectrum, contrary to the binList which has (itemCount -
		// 1) bins. This is because when populating binMassSpectrum in the
		// setupBinnability() function, the smallest mz value found in all the
		// mass spectra of the MassSpectrumList is first seeded as such, and is
		// thus the first mass peak item of the mass spectrum. Then, for each bin,
		// a new mass peak is appended to the spectrum. At the end, there are bins
		// + 1 mass peaks in the spectrum, which is identical to itemCount.

		REQUIRE(binMassSpectrum.size() == itemCount);

		// Prepare a copy spectrum for later comparisons.

		MassSpectrum otherBinMassSpectrum;

		res = otherBinMassSpectrum.setupBinnability(massSpectra);
		REQUIRE(res == true);

		SECTION("Combine (in a binned manner) a mass peak into "
				"the mass spectrum (combination occurs roughly in the "
				"middle of the spectrum", "[MassSpectrum]")
		{

			// Generate a random index
			std::default_random_engine generator;
			std::uniform_int_distribution<int>distribution(10, itemCount - 10);
			int mzIndex = distribution(generator);
			//cout << "Random index: " << mzIndex << "\n";

			double origPrevBinMz = binMassSpectrum.at(mzIndex)->mz();
			double origPrevBinI = binMassSpectrum.at(mzIndex)->i();

			double origNextBinMz = binMassSpectrum.at(mzIndex + 1)->mz();
			double origNextBinI = binMassSpectrum.at(mzIndex + 1)->i();

			double origBinInterval = origNextBinMz - origPrevBinMz;

			//cout << std::setprecision(15) << "orig prev bin: (" << origPrevBinMz << "," << origPrevBinI  << ")\n";
			//cout << std::setprecision(15) << "orig next bin: (" << origNextBinMz << "," << origNextBinI  << ")\n";
			//cout << std::setprecision(15) << "orig bin interval: " << origBinInterval << "\n";

			// Create a new MassPeak having an mz value that sits between both bins.
			// It should thus be set to origPrevMz.

			MassPeak newMassPeak(origPrevBinMz + (origBinInterval / 2), 1234);

			binMassSpectrum.combineBinned(newMassPeak);

			// The only item that should have changed is the bin at mzIndex
			// (origPrevBinMz).

			int failedComparisons = 0;

			for(int iter = 0; iter < itemCount; ++iter)
			{
				double mz = binMassSpectrum.at(iter)->mz();
				double i = binMassSpectrum.at(iter)->i();

				double otherMz = otherBinMassSpectrum.at(iter)->mz();
				double otherI = otherBinMassSpectrum.at(iter)->i();

				if(iter == mzIndex)
				{
					REQUIRE(i == 1234);
					continue;
				}

				if(mz != otherMz || i != otherI)
					++failedComparisons;
			}

			REQUIRE(failedComparisons == 0);

			delete massSpectra.at(1);
			delete massSpectra.at(0);

			massSpectra.clear();

			// At this point we have tested the binned combination of a mass peak into
			// the inner part of a spectrum (not at the first or at the last bin of
			// the spectrum).
		}

		SECTION("Combine (in a binned manner) a mass peak into "
				"the mass spectrum (combination occurs right at top).", "[MassSpectrum]")
		{

			int mzIndex = 0;
			//cout << "First bin index: " << mzIndex << "\n";

			double origPrevBinMz = binMassSpectrum.at(mzIndex)->mz();
			double origPrevBinI = binMassSpectrum.at(mzIndex)->i();

			double origNextBinMz = binMassSpectrum.at(mzIndex + 1)->mz();
			double origNextBinI = binMassSpectrum.at(mzIndex + 1)->i();

			double origBinInterval = origNextBinMz - origPrevBinMz;

			//cout << std::setprecision(15) << "orig prev bin: (" << origPrevBinMz << "," << origPrevBinI  << ")\n";
			//cout << std::setprecision(15) << "orig next bin: (" << origNextBinMz << "," << origNextBinI  << ")\n";
			//cout << std::setprecision(15) << "orig bin interval: " << origBinInterval << "\n";
			//cout << std::setprecision(15) << "new mz value: " << origPrevBinMz << "\n";

			// Create a new MassPeak having an mz value that is identical to the one
			// of the first bin. Note that it is illegal to provide a mz value less
			// than the mz value of the first bin, because that would means that
			// setupBinnability() had not be called properly.

			MassPeak newMassPeak(origPrevBinMz, 1234);

			binMassSpectrum.combineBinned(newMassPeak);

			// The only item that should have changed is the bin at mzIndex, that is
			// the very first mz value has changed and is the one combined.

			REQUIRE(binMassSpectrum.at(mzIndex)->mz() == origPrevBinMz);
			REQUIRE(binMassSpectrum.at(mzIndex)->i() == origPrevBinI + 1234);

			// Now, skipping the first mass peak of binMassSpectrum, check the
			// remaining elements of the spectra. They should be identical. Note that
			// there is not dephased iteration, because when combining in binned mode,
			// the number of bins never changes.

			int failedComparisons = 0;

			for(int iter = 1; iter < itemCount; ++iter)
			{
				double mz = binMassSpectrum.at(iter)->mz();
				double i = binMassSpectrum.at(iter)->i();

				double otherMz = otherBinMassSpectrum.at(iter)->mz();
				double otherI = otherBinMassSpectrum.at(iter)->i();

				if(mz != otherMz || i != otherI)
					++failedComparisons;
			}

			REQUIRE(failedComparisons == 0);

			delete massSpectra.at(1);
			delete massSpectra.at(0);

			massSpectra.clear();

			// At this point we have tested the binned combination of a mass peak as
			// the new top mass peak of a spectrum.
		}

		SECTION("Combine (in a binned manner) a mass peak into "
				"the mass spectrum (combination occurs right at bottom).", "[MassSpectrum]")
		{

			int mzIndex = itemCount - 1;
			//cout << "First bin index: " << mzIndex << "\n";

			double origPrevBinMz = binMassSpectrum.at(mzIndex)->mz();
			double origPrevBinI = binMassSpectrum.at(mzIndex)->i();

			double origPrevPrevBinMz = binMassSpectrum.at(mzIndex - 1)->mz();
			double origPrevPrevBinI = binMassSpectrum.at(mzIndex - 1)->i();

			double origBinInterval = origPrevPrevBinMz - origPrevBinMz;

			//cout << std::setprecision(15) << "orig prev bin: (" << origPrevBinMz << "," << origPrevBinI  << ")\n";
			//cout << std::setprecision(15) << "orig prev prev bin: (" << origPrevPrevBinMz << "," << origPrevPrevBinI  << ")\n";
			//cout << std::setprecision(15) << "orig bin interval: " << origBinInterval << "\n";
			//cout << std::setprecision(15) << "new mz value: " << origPrevBinMz << "\n";

			// Create a new MassPeak having an mz value that is identical to the one
			// of the last bin. 
			MassPeak newMassPeak(origPrevBinMz, 1234);

			binMassSpectrum.combineBinned(newMassPeak);

			// The only item that should have changed is the bin at mzIndex, that is
			// the very last mz value has changed and is the one combined.

			REQUIRE(binMassSpectrum.at(mzIndex)->mz() == origPrevBinMz);
			REQUIRE(binMassSpectrum.at(mzIndex)->i() == origPrevBinI + 1234);

			// Now, skipping the last mass peak of binMassSpectrum, check the
			// remaining elements of the spectra. They should be identical. Note that
			// there is not dephased iteration, because when combining in binned mode,
			// the number of bins never changes.

			int failedComparisons = 0;

			for(int iter = 1; iter < itemCount - 1; ++iter)
			{
				double mz = binMassSpectrum.at(iter)->mz();
				double i = binMassSpectrum.at(iter)->i();

				double otherMz = otherBinMassSpectrum.at(iter)->mz();
				double otherI = otherBinMassSpectrum.at(iter)->i();

				if(mz != otherMz || i != otherI)
					++failedComparisons;
			}

			REQUIRE(failedComparisons == 0);

			delete massSpectra.at(1);
			delete massSpectra.at(0);

			massSpectra.clear();

			// At this point we have tested the binned combination of a mass peak as
			// the new top mass peak of a spectrum.
		}
	}


	TEST_CASE("Combination of mass spectrum list into a mass " 
			"spectrum (binnable case).", "[MassSpectrum]") 
	{
		// Preparation of various binnable spectra (conventional qTOF experiments on
		// proteins (see tests/data/xy/protein-microtofq-bruker.mzml-spec-index-<0--4>.xy)

		QStringList xyFormatStringList;
		xyFormatStringList
			<< "694.5166264003 133.0000000000"
			<< "694.5333824250 195.0000000000"
			<< "694.5501386518 225.0000000000"
			<< "694.5668950807 220.0000000000"
			<< "694.5836517118 286.0000000000"
			<< "694.6004085450 332.0000000000"
			<< "694.6171655803 282.0000000000"
			<< "694.6339228177 164.0000000000"
			<< "694.6506802573 140.0000000000"
			<< "694.6674378989 156.0000000000"
			<< "694.6841957428 153.0000000000"
			<< "694.7009537887 89.0000000000"
			<< "694.7177120368 104.0000000000"
			<< "694.7344704869 143.0000000000"
			<< "694.7512291393 196.0000000000"
			<< "694.7679879937 295.0000000000"
			<< "694.7847470503 396.0000000000"
			<< "694.8015063090 399.0000000000"
			<< "694.8182657698 369.0000000000"
			<< "694.8350254327 345.0000000000" ;

		int xyFormatStringListSize = xyFormatStringList.size();
		QString xyFormatString = xyFormatStringList.join("\n");
		MassSpectrum *massSpecIdx0 = new MassSpectrum;
		massSpecIdx0->initialize(xyFormatString);
		REQUIRE(xyFormatStringListSize == massSpecIdx0->size());

		xyFormatStringList.clear();
		xyFormatStringList
			<< "694.5166130065 186.0000000000"
			<< "694.5333690308 264.0000000000"
			<< "694.5501252573 288.0000000000"
			<< "694.5668816859 264.0000000000"
			<< "694.5836383166 276.0000000000"
			<< "694.6003951495 303.0000000000"
			<< "694.6171521845 267.0000000000"
			<< "694.6339094216 189.0000000000"
			<< "694.6506668608 170.0000000000"
			<< "694.6674245022 165.0000000000"
			<< "694.6841823457 173.0000000000"
			<< "694.7009403913 159.0000000000"
			<< "694.7176986390 180.0000000000"
			<< "694.7344570889 212.0000000000"
			<< "694.7512157409 245.0000000000"
			<< "694.7679745950 299.0000000000"
			<< "694.7847336512 357.0000000000"
			<< "694.8014929096 381.0000000000"
			<< "694.8182523701 343.0000000000"
			<< "694.8350120327 285.0000000000" ;

		xyFormatStringListSize = xyFormatStringList.size();
		xyFormatString = xyFormatStringList.join("\n");
		MassSpectrum *massSpecIdx1 = new MassSpectrum;
		massSpecIdx1->initialize(xyFormatString);
		REQUIRE(xyFormatStringListSize == massSpecIdx1->size());

		xyFormatStringList.clear();
		xyFormatStringList 
			<< "694.5166130065 204.0000000000"
			<< "694.5333690308 277.0000000000"
			<< "694.5501252573 300.0000000000"
			<< "694.5668816859 286.0000000000"
			<< "694.5836383166 273.0000000000"
			<< "694.6003951495 276.0000000000"
			<< "694.6171521845 268.0000000000"
			<< "694.6339094216 235.0000000000"
			<< "694.6506668608 223.0000000000"
			<< "694.6674245022 195.0000000000"
			<< "694.6841823457 220.0000000000"
			<< "694.7009403913 233.0000000000"
			<< "694.7176986390 249.0000000000"
			<< "694.7344570889 247.0000000000"
			<< "694.7512157409 263.0000000000"
			<< "694.7679745950 299.0000000000"
			<< "694.7847336512 335.0000000000"
			<< "694.8014929096 354.0000000000"
			<< "694.8182523701 317.0000000000"
			<< "694.8350120327 249.0000000000" ;

		xyFormatStringListSize = xyFormatStringList.size();
		xyFormatString = xyFormatStringList.join("\n");
		MassSpectrum *massSpecIdx2 = new MassSpectrum;
		massSpecIdx2->initialize(xyFormatString);
		REQUIRE(xyFormatStringListSize == massSpecIdx2->size());

		xyFormatStringList.clear();
		xyFormatStringList
			<< "694.5165997465 215.0000000000"
			<< "694.5333557706 264.0000000000"
			<< "694.5501119967 292.0000000000"
			<< "694.5668684250 285.0000000000"
			<< "694.5836250554 296.0000000000"
			<< "694.6003818880 310.0000000000"
			<< "694.6171389226 304.0000000000"
			<< "694.6338961594 256.0000000000"
			<< "694.6506535983 232.0000000000"
			<< "694.6674112394 212.0000000000"
			<< "694.6841690826 235.0000000000"
			<< "694.7009271278 255.0000000000"
			<< "694.7176853753 284.0000000000"
			<< "694.7344438248 268.0000000000"
			<< "694.7512024765 272.0000000000"
			<< "694.7679613303 304.0000000000"
			<< "694.7847203862 316.0000000000"
			<< "694.8014796442 304.0000000000"
			<< "694.8182391044 288.0000000000"
			<< "694.8349987667 250.0000000000" ;

		xyFormatStringListSize = xyFormatStringList.size();
		xyFormatString = xyFormatStringList.join("\n");
		MassSpectrum *massSpecIdx3 = new MassSpectrum;
		massSpecIdx3->initialize(xyFormatString);
		REQUIRE(xyFormatStringListSize == massSpecIdx3->size());

		xyFormatStringList.clear();
		xyFormatStringList
			<< "694.5165997465 237.0000000000"
			<< "694.5333557706 261.0000000000"
			<< "694.5501119967 304.0000000000"
			<< "694.5668684250 312.0000000000"
			<< "694.5836250554 297.0000000000"
			<< "694.6003818880 279.0000000000"
			<< "694.6171389226 288.0000000000"
			<< "694.6338961594 268.0000000000"
			<< "694.6506535983 269.0000000000"
			<< "694.6674112394 238.0000000000"
			<< "694.6841690826 248.0000000000"
			<< "694.7009271278 242.0000000000"
			<< "694.7176853753 250.0000000000"
			<< "694.7344438248 256.0000000000"
			<< "694.7512024765 266.0000000000"
			<< "694.7679613303 276.0000000000"
			<< "694.7847203862 270.0000000000"
			<< "694.8014796442 269.0000000000"
			<< "694.8182391044 281.0000000000"
			<< "694.8349987667 262.0000000000" ;

		xyFormatStringListSize = xyFormatStringList.size();
		xyFormatString = xyFormatStringList.join("\n");
		MassSpectrum *massSpecIdx4 = new MassSpectrum;
		massSpecIdx4->initialize(xyFormatString);
		REQUIRE(xyFormatStringListSize == massSpecIdx4->size());

		xyFormatStringList.clear();

		SECTION("Combination (binned case) with no mz range.", "[MassSpectrum]")
		{
			// We now have five mass spectra. We can now construct the mass spectrum
			// list.
			MassSpectrumList massSpectra;
			massSpectra << massSpecIdx0 << massSpecIdx1 << massSpecIdx2 << massSpecIdx3 << massSpecIdx4;

			// The combined mass spectrum as computed apart:

			xyFormatStringList
				<< "694.5165997465 1500.0000000000"
				<< "694.5333557712 1332.0000000000"
				<< "694.5501119980 1410.0000000000"
				<< "694.5668684269 1363.0000000000"
				<< "694.5836250580 1424.0000000000"
				<< "694.6003818912 1503.0000000000"
				<< "694.6171389265 1341.0000000000"
				<< "694.6338961639 1089.0000000000"
				<< "694.6506536035 983.0000000000"
				<< "694.6674112451 999.0000000000"
				<< "694.6841690890 1043.0000000000"
				<< "694.7009271349 1015.0000000000"
				<< "694.7176853830 1057.0000000000"
				<< "694.7344438331 1140.0000000000"
				<< "694.7512024855 1284.0000000000"
				<< "694.7679613399 1479.0000000000"
				<< "694.7847203965 1661.0000000000"
				<< "694.8014796552 1703.0000000000"
				<< "694.8182391160 1541.0000000000"
				<< "694.8349987789 879.0000000000" ;

			xyFormatStringListSize = xyFormatStringList.size();
			xyFormatString = xyFormatStringList.join("\n");
			MassSpectrum theorCombinedMassSpectrum;
			theorCombinedMassSpectrum.initialize(xyFormatString);
			REQUIRE(xyFormatStringListSize == theorCombinedMassSpectrum.size());

			xyFormatStringList.clear();

			// Finally actually test the combination function.

			MassSpectrum combinedMassSpectrum;
			combinedMassSpectrum.combine(massSpectra);

			// Check that both the theor and the actual combined spectra have the same
			// number of mass peak instances.

			REQUIRE(theorCombinedMassSpectrum.size() == combinedMassSpectrum.size());

			// Now test that the obained mass combined mass spectrum is identical to the
			// one computed apart.

			int failedComparisons = 0;

			for(int iter = 0; iter < combinedMassSpectrum.size(); ++iter)
			{
				double mz = combinedMassSpectrum.at(iter)->mz();
				double i = combinedMassSpectrum.at(iter)->i();

				double theorMz = theorCombinedMassSpectrum.at(iter)->mz();
				double theorI = theorCombinedMassSpectrum.at(iter)->i();

				//cout << std::setprecision(15)
				//<< "(" << mz << "," << i << ") vs (" << theorMz << "," << theorI << ")\n";

				// It is important to use the almostEqual() because otherwise we get
				// spurious messages.
				if(!msXpS::almostEqual(mz,theorMz, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << mz << " vs theor " << theorMz << "\n";

					++failedComparisons;
				}

				if(!msXpS::almostEqual(i,theorI, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << i << " vs theor " << theorI << "\n";

					++failedComparisons;
				}
			}

			REQUIRE(failedComparisons == 0);
		}

		SECTION("Combination (binned case) with mz range.", "[MassSpectrum]")
		{
			// We now have five mass spectra. We can now construct the mass spectrum
			// list.
			MassSpectrumList massSpectra;
			massSpectra << massSpecIdx0 << massSpecIdx1 << massSpecIdx2 << massSpecIdx3 << massSpecIdx4;

			// We need actually to prepare mzStart and mzEnd the bins as they are
			// going to be prepared in the combined mass spectrum:

			//"694.5165997465 0"
			//"694.5333557712 0"
			//"694.5501119980 0"
			//"694.5668684269 0"
			//"694.5836250580 0"
			//"694.6003818912 0"
			//"694.6171389265 0"
			//"694.6338961639 0"
			//"694.6506536035 0"
			//"694.6674112451 0"
			//"694.6841690890 0"
			//"694.7009271349 0"
			//"694.7176853830 0"
			//"694.7344438331 0"
			//"694.7512024855 0"
			//"694.7679613399 0"
			//"694.7847203965 0"
			//"694.8014796552 0"
			//"694.8182391160 0"
			//"694.8349987789 0"

			// Skip 5 spectra on the top (massSpecIdx0).
			// A mass peak is in the range if mz >= mzStart.
			double mzStart = 694.6003818912;

			// Skip 5 spectra on the bottom
			// A mass peak is in the range if mz <= mzEnd.
			double mzEnd = 694.760000000;

			//cout << std::setprecision(15) << "Combining spectra with mz range: ["
			//<< mzStart << "," << mzEnd << "]\n";

			// The combined mass spectrum as computed apart:

			xyFormatStringList
				<< "694.5165997465 0.0000000000"
				<< "694.5333557712 0.0000000000"
				<< "694.5501119980 0.0000000000"
				<< "694.5668684269 0.0000000000"
				<< "694.5836250580 0.0000000000"
				<< "694.6003818912 1503.0000000000"
				<< "694.6171389265 1341.0000000000"
				<< "694.6338961639 1089.0000000000"
				<< "694.6506536035 983.0000000000"
				<< "694.6674112451 999.0000000000"
				<< "694.6841690890 1043.0000000000"
				<< "694.7009271349 1015.0000000000"
				<< "694.7176853830 1057.0000000000"
				<< "694.7344438331 1140.0000000000"
				<< "694.7512024855 704.0000000000"
				<< "694.7679613399 0.0000000000"
				<< "694.7847203965 0.0000000000"
				<< "694.8014796552 0.0000000000"
				<< "694.8182391160 0.0000000000"
				<< "694.8349987789 0.0000000000";

			xyFormatStringListSize = xyFormatStringList.size();
			xyFormatString = xyFormatStringList.join("\n");
			MassSpectrum theorCombinedMassSpectrum;
			theorCombinedMassSpectrum.initialize(xyFormatString);
			REQUIRE(xyFormatStringListSize == theorCombinedMassSpectrum.size());

			xyFormatStringList.clear();

			// Finally actually test the combination function.

			MassSpectrum combinedMassSpectrum;

			combinedMassSpectrum.combine(massSpectra, mzStart, mzEnd);

			// Check that both the theor and the actual combined spectra have the same
			// number of mass peak instances.

			REQUIRE(theorCombinedMassSpectrum.size() == combinedMassSpectrum.size());

			// Now test that the obained mass combined mass spectrum is identical to the
			// one computed apart.

			int failedComparisons = 0;

			for(int iter = 0; iter < combinedMassSpectrum.size(); ++iter)
			{
				double mz = combinedMassSpectrum.at(iter)->mz();
				double i = combinedMassSpectrum.at(iter)->i();

				double theorMz = theorCombinedMassSpectrum.at(iter)->mz();
				double theorI = theorCombinedMassSpectrum.at(iter)->i();

				//cout << std::setprecision(15)
				//<< "(" << mz << "," << i << ") vs (" << theorMz << "," << theorI << ")\n";

				// It is important to use the almostEqual() because otherwise we get
				// spurious messages.
				if(!msXpS::almostEqual(mz,theorMz, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << mz << " vs theor " << theorMz << "\n";

					++failedComparisons;
				}

				if(!msXpS::almostEqual(i,theorI, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << i << " vs theor " << theorI << "\n";

					++failedComparisons;
				}
			}

			REQUIRE(failedComparisons == 0);
		}

		SECTION("Combination (binned case) of mzList/iList.", "[MassSpectrum]")
		{
			// We now have five mass spectra. We can now construct the mass spectrum
			// list. Only use the first four, as the fifth will be combined in the form
			// of mzList and iList, which allows use to recycle the combined spectrum
			// form an earlier test.

			MassSpectrumList massSpectra;
			massSpectra << massSpecIdx0 << massSpecIdx1 << massSpecIdx2 << massSpecIdx3;

			// Prepare mzList and iList to make the combination later.

			QList<double> mzList;
			mzList = {694.5165997465,694.5333557706,694.5501119967,694.5668684250,694.5836250554,694.6003818880,694.6171389226,694.6338961594,694.6506535983,694.6674112394,694.6841690826,694.7009271278,694.7176853753,694.7344438248,694.7512024765,694.7679613303,694.7847203862,694.8014796442,694.8182391044,694.8349987667};

			QList<double> iList = {237.0,261.0,304.0,312.0,297.0,279.0,288.0,268.0,269.0,238.0,248.0,242.0,250.0,256.0,266.0,276.0,270.0,269.0,281.0,262.0};

			int itemCount = mzList.size();
			REQUIRE(itemCount == iList.size());

			MassSpectrum combinedMassSpectrum;
			combinedMassSpectrum.combine(massSpectra);
			// Doing the last combination like the following works.
			// combinedMassSpectrum.combine(*massSpecIdx4);

			combinedMassSpectrum.combine(mzList, iList);

			// The combined mass spectrum as computed apart (see earlier)

			QStringList xyFormatStringList;
			xyFormatStringList
				<< "694.5165997465 1500.0000000000"
				<< "694.5333557712 1332.0000000000"
				<< "694.5501119980 1410.0000000000"
				<< "694.5668684269 1363.0000000000"
				<< "694.5836250580 1424.0000000000"
				<< "694.6003818912 1503.0000000000"
				<< "694.6171389265 1341.0000000000"
				<< "694.6338961639 1089.0000000000"
				<< "694.6506536035 983.0000000000"
				<< "694.6674112451 999.0000000000"
				<< "694.6841690890 1043.0000000000"
				<< "694.7009271349 1015.0000000000"
				<< "694.7176853830 1057.0000000000"
				<< "694.7344438331 1140.0000000000"
				<< "694.7512024855 1284.0000000000"
				<< "694.7679613399 1479.0000000000"
				<< "694.7847203965 1661.0000000000"
				<< "694.8014796552 1703.0000000000"
				<< "694.8182391160 1541.0000000000"
				<< "694.8349987789 879.0000000000" ;

			xyFormatStringListSize = xyFormatStringList.size();
			xyFormatString = xyFormatStringList.join("\n");
			MassSpectrum theorCombinedMassSpectrum;
			theorCombinedMassSpectrum.initialize(xyFormatString);
			REQUIRE(xyFormatStringListSize == theorCombinedMassSpectrum.size());

			// Check that both the theor and the actual combined spectra have the same
			// number of mass peak instances.

			REQUIRE(theorCombinedMassSpectrum.size() == combinedMassSpectrum.size());

			// Now test that the obained mass combined mass spectrum is identical to the
			// one computed apart.

			int failedComparisons = 0;

			for(int iter = 0; iter < combinedMassSpectrum.size(); ++iter)
			{
				double mz = combinedMassSpectrum.at(iter)->mz();
				double i = combinedMassSpectrum.at(iter)->i();

				double theorMz = theorCombinedMassSpectrum.at(iter)->mz();
				double theorI = theorCombinedMassSpectrum.at(iter)->i();

				//cout << std::setprecision(15)
				//<< "(" << mz << "," << i << ") vs (" << theorMz << "," << theorI << ")\n";

				// It is important to use the almostEqual() because otherwise we get
				// spurious messages.
				if(!msXpS::almostEqual(mz,theorMz, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << mz << " vs theor " << theorMz << "\n";

					++failedComparisons;
				}

				if(!msXpS::almostEqual(i,theorI, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << i << " vs theor " << theorI << "\n";

					++failedComparisons;
				}
			}

			REQUIRE(failedComparisons == 0);
		}

		SECTION("Combination (binned case) of mzList/iList with mz range.", "[MassSpectrum]")
		{
			// We now have five mass spectra. We can now construct the mass spectrum
			// list. Only use the first four, as the fifth will be combined in the form
			// of mzList and iList, which allows use to recycle the combined spectrum
			// form an earlier test.

			MassSpectrumList massSpectra;
			massSpectra << massSpecIdx0 << massSpecIdx1 << massSpecIdx2 << massSpecIdx3;

			// Prepare mzList and iList to make the combination later.

			// The mzList and iList are the lists used earlier to craft massSpecIdx0.

			QList<double> mzList;
			mzList = {694.5165997465,694.5333557706,694.5501119967,694.5668684250,694.5836250554,694.6003818880,694.6171389226,694.6338961594,694.6506535983,694.6674112394,694.6841690826,694.7009271278,694.7176853753,694.7344438248,694.7512024765,694.7679613303,694.7847203862,694.8014796442,694.8182391044,694.8349987667};

			QList<double> iList = {237.0,261.0,304.0,312.0,297.0,279.0,288.0,268.0,269.0,238.0,248.0,242.0,250.0,256.0,266.0,276.0,270.0,269.0,281.0,262.0};

			int itemCount = mzList.size();
			REQUIRE(itemCount == iList.size());

			// A mass peak is in the range if mz >= mzStart.
			double mzStart = 694.6003818912;

			// Skip 5 spectra on the bottom
			// A mass peak is in the range if mz <= mzEnd.
			double mzEnd = 694.760000000;

			//cout << std::setprecision(15) << "Combining spectra with mz range: ["
			//<< mzStart << "," << mzEnd << "]\n";

			MassSpectrum combinedMassSpectrum;
			combinedMassSpectrum.combine(massSpectra, mzStart, mzEnd);
			// Doing the last combination like the following works.
			// combinedMassSpectrum.combine(*massSpecIdx4, mzStart, mzEnd);

			combinedMassSpectrum.combine(mzList, iList, mzStart, mzEnd);

			// The combined mass spectrum as computed apart (see earlier)

			xyFormatStringList
				<< "694.5165997465 0.0000000000"
				<< "694.5333557712 0.0000000000"
				<< "694.5501119980 0.0000000000"
				<< "694.5668684269 0.0000000000"
				<< "694.5836250580 0.0000000000"
				<< "694.6003818912 1503.0000000000"
				<< "694.6171389265 1341.0000000000"
				<< "694.6338961639 1089.0000000000"
				<< "694.6506536035 983.0000000000"
				<< "694.6674112451 999.0000000000"
				<< "694.6841690890 1043.0000000000"
				<< "694.7009271349 1015.0000000000"
				<< "694.7176853830 1057.0000000000"
				<< "694.7344438331 1140.0000000000"
				<< "694.7512024855 704.0000000000"
				<< "694.7679613399 0.0000000000"
				<< "694.7847203965 0.0000000000"
				<< "694.8014796552 0.0000000000"
				<< "694.8182391160 0.0000000000"
				<< "694.8349987789 0.0000000000";

			xyFormatStringListSize = xyFormatStringList.size();
			xyFormatString = xyFormatStringList.join("\n");
			MassSpectrum theorCombinedMassSpectrum;
			theorCombinedMassSpectrum.initialize(xyFormatString);
			REQUIRE(xyFormatStringListSize == theorCombinedMassSpectrum.size());

			// Check that both the theor and the actual combined spectra have the same
			// number of mass peak instances.

			REQUIRE(theorCombinedMassSpectrum.size() == combinedMassSpectrum.size());

			// Now test that the obained mass combined mass spectrum is identical to the
			// one computed apart.

			int failedComparisons = 0;

			for(int iter = 0; iter < combinedMassSpectrum.size(); ++iter)
			{
				double mz = combinedMassSpectrum.at(iter)->mz();
				double i = combinedMassSpectrum.at(iter)->i();

				double theorMz = theorCombinedMassSpectrum.at(iter)->mz();
				double theorI = theorCombinedMassSpectrum.at(iter)->i();

				//cout << std::setprecision(15)
				//<< "(" << mz << "," << i << ") vs (" << theorMz << "," << theorI << ")\n";

				// It is important to use the almostEqual() because otherwise we get
				// spurious messages.
				if(!msXpS::almostEqual(mz,theorMz, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << mz << " vs theor " << theorMz << "\n";

					++failedComparisons;
				}

				if(!msXpS::almostEqual(i,theorI, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << i << " vs theor " << theorI << "\n";

					++failedComparisons;
				}
			}

			REQUIRE(failedComparisons == 0);
		}


		delete massSpecIdx0;
		delete massSpecIdx1;
		delete massSpecIdx2;
		delete massSpecIdx3;
		delete massSpecIdx4; 

		// At this point we have tested the binned combination into an empty mass
		// spectrum of a mass spectrum list containing five mass spectra.
	}


	TEST_CASE("Combination of mass spectrum list into a mass " 
			"spectrum (non-binnable case).", "[MassSpectrum]") 
	{
		// Preparation of various mobility-experiment data that are not binnable
		// mass spectra because the mz intervals are not equal all along the mz
		// vector. First twenty mass peaks from files
		// tests/data/xy/protein-mobility-waters-synapt-hdms.mzml-spec-index-[0--4].xy
		// and protein-mobility-waters-synapt-hdms.mzml-combination-spec-idx-0-to-4.xy.

		QStringList xyFormatStringList;
		xyFormatStringList
			<< "2207.9724474677 0.0000000000"
			<< "2207.9932149877 3.0000000000"
			<< "2208.0137381840 32.0000000000"
			<< "2208.0342613806 0.0000000000";

		int xyFormatStringListSize = xyFormatStringList.size();
		QString xyFormatString = xyFormatStringList.join("\n");
		MassSpectrum *massSpecIdx0 = new MassSpectrum;
		massSpecIdx0->initialize(xyFormatString);
		REQUIRE(xyFormatStringListSize == massSpecIdx0->size());

		xyFormatStringList.clear();
		xyFormatStringList
			<< "1861.6054557607 0.0000000000"
			<< "1861.6243881557 4.0000000000"
			<< "1861.6431984064 16.0000000000"
			<< "1861.6620086572 0.0000000000";

		xyFormatStringListSize = xyFormatStringList.size();
		xyFormatString = xyFormatStringList.join("\n");
		MassSpectrum *massSpecIdx1 = new MassSpectrum;
		massSpecIdx1->initialize(xyFormatString);
		REQUIRE(xyFormatStringListSize == massSpecIdx1->size());

		xyFormatStringList.clear();
		xyFormatStringList 
			<< "2173.2127519882 0.0000000000"
			<< "2173.2330305731 52.0000000000"
			<< "2173.2533091582 3.0000000000"
			<< "2173.2738320637 0.0000000000";

		xyFormatStringListSize = xyFormatStringList.size();
		xyFormatString = xyFormatStringList.join("\n");
		MassSpectrum *massSpecIdx2 = new MassSpectrum;
		massSpecIdx2->initialize(xyFormatString);
		REQUIRE(xyFormatStringListSize == massSpecIdx2->size());

		xyFormatStringList.clear();
		xyFormatStringList
			<< "1879.3452969428 0.0000000000"
			<< "1879.3642294765 27.0000000000"
			<< "1879.3831620103 0.0000000000";

		xyFormatStringListSize = xyFormatStringList.size();
		xyFormatString = xyFormatStringList.join("\n");
		MassSpectrum *massSpecIdx3 = new MassSpectrum;
		massSpecIdx3->initialize(xyFormatString);
		REQUIRE(xyFormatStringListSize == massSpecIdx3->size());

		xyFormatStringList.clear();
		xyFormatStringList
			<< "1848.9427702443 0.0000000000"
			<< "1848.9615803966 21.0000000000"
			<< "1848.9803905489 17.0000000000"
			<< "1848.9990785576 0.0000000000"
			<< "2007.9375013022 0.0000000000"
			<< "2007.9570456000 34.0000000000"
			<< "2007.9765898979 3.0000000000"
			<< "2007.9962563479 0.0000000000";

		xyFormatStringListSize = xyFormatStringList.size();
		xyFormatString = xyFormatStringList.join("\n");
		MassSpectrum *massSpecIdx4 = new MassSpectrum;
		massSpecIdx4->initialize(xyFormatString);
		REQUIRE(xyFormatStringListSize == massSpecIdx4->size());

		xyFormatStringList.clear();


		SECTION("Combination (not-binned case) with no mz range.", "[MassSpectrum]")
		{
			// We now have five mass spectra. We can now construct the mass spectrum
			// list.

			MassSpectrumList massSpectra;
			massSpectra << massSpecIdx0 << massSpecIdx1 << massSpecIdx2 << massSpecIdx3 << massSpecIdx4;

			// The combined mass spectrum as computed apart:
			xyFormatStringList.clear();
			xyFormatStringList
				<< "1848.9427702443 0.0000000000"
				<< "1848.9615803966 21.0000000000"
				<< "1848.9803905489 17.0000000000"
				<< "1848.9990785576 0.0000000000"
				<< "1861.6054557607 0.0000000000"
				<< "1861.6243881557 4.0000000000"
				<< "1861.6431984064 16.0000000000"
				<< "1861.6620086572 0.0000000000"
				<< "1879.3452969428 0.0000000000"
				<< "1879.3642294765 27.0000000000"
				<< "1879.3831620103 0.0000000000"
				<< "2007.9375013022 0.0000000000"
				<< "2007.9570456000 34.0000000000"
				<< "2007.9765898979 3.0000000000"
				<< "2007.9962563479 0.0000000000"
				<< "2173.2127519882 0.0000000000"
				<< "2173.2330305731 52.0000000000"
				<< "2173.2533091582 3.0000000000"
				<< "2173.2738320637 0.0000000000"
				<< "2207.9724474677 0.0000000000"
				<< "2207.9932149877 3.0000000000"
				<< "2208.0137381840 32.0000000000"
				<< "2208.0342613806 0.0000000000";

			xyFormatStringListSize = xyFormatStringList.size();
			xyFormatString = xyFormatStringList.join("\n");
			MassSpectrum theorCombinedMassSpectrum;
			theorCombinedMassSpectrum.initialize(xyFormatString);
			REQUIRE(xyFormatStringListSize == theorCombinedMassSpectrum.size());

			xyFormatStringList.clear();

			// Finally actually test the combination function.

			MassSpectrum combinedMassSpectrum;
			combinedMassSpectrum.combine(massSpectra);

			// Check that both the theor and the actual combined spectra have the same
			// number of mass peak instances.

			REQUIRE(theorCombinedMassSpectrum.size() == combinedMassSpectrum.size());

			// Now test that the obained mass combined mass spectrum is identical to the
			// one computed apart.

			int failedComparisons = 0;

			for(int iter = 0; iter < combinedMassSpectrum.size(); ++iter)
			{
				double mz = combinedMassSpectrum.at(iter)->mz();
				double i = combinedMassSpectrum.at(iter)->i();

				double theorMz = theorCombinedMassSpectrum.at(iter)->mz();
				double theorI = theorCombinedMassSpectrum.at(iter)->i();

				//cout << std::setprecision(15)
				//<< "(" << mz << "," << i << ") vs (" << theorMz << "," << theorI << ")\n";

				// It is important to use the almostEqual() because otherwise we get
				// spurious messages.
				if(!msXpS::almostEqual(mz,theorMz, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << mz << " vs theor " << theorMz << "\n";

					++failedComparisons;
				}

				if(!msXpS::almostEqual(i,theorI, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << i << " vs theor " << theorI << "\n";

					++failedComparisons;
				}
			}

			REQUIRE(failedComparisons == 0);
		}

		SECTION("Combine (not-binned case) with mz range.", "[MassSpectrum]")
		{
			// We now have five mass spectra. We can now construct the mass spectrum
			// list.

			MassSpectrumList massSpectra;
			massSpectra << massSpecIdx0 << massSpecIdx1 << massSpecIdx2 << massSpecIdx3 << massSpecIdx4;

			// The un-mz-ranged combination above provides the spectrum below 
			//
			//"1848.9427702443 0.0000000000"
			//"1848.9615803966 21.0000000000"
			//"1848.9803905489 17.0000000000"
			//"1848.9990785576 0.0000000000"
			//"1861.6054557607 0.0000000000"
			//"1861.6243881557 4.0000000000"
			//"1861.6431984064 16.0000000000"
			//"1861.6620086572 0.0000000000"
			//"1879.3452969428 0.0000000000"
			//"1879.3642294765 27.0000000000"
			//"1879.3831620103 0.0000000000"
			//"2007.9375013022 0.0000000000"
			//"2007.9570456000 34.0000000000"
			//"2007.9765898979 3.0000000000"
			//"2007.9962563479 0.0000000000"
			//"2173.2127519882 0.0000000000"
			//"2173.2330305731 52.0000000000"
			//"2173.2533091582 3.0000000000"
			//"2173.2738320637 0.0000000000"
			//"2207.9724474677 0.0000000000"
			//"2207.9932149877 3.0000000000"
			//"2208.0137381840 32.0000000000"
			//"2208.0342613806 0.0000000000";

			// Skip 5 spectra on the top (massSpecIdx0).
			// A mass peak is in the range if mz >= mzStart.
			double mzStart = 1861.61;

			// Skip 5 spectra on the bottom
			// A mass peak is in the range if mz <= mzEnd.
			double mzEnd = 2173.26;

			//cout << std::setprecision(15) << "Combining spectra with mz range: ["
			//<< mzStart << "," << mzEnd << "]\n";

			// The combined mass spectrum as computed apart:

			xyFormatStringList.clear();
			xyFormatStringList
				<< "1861.6243881557 4.0000000000"
				<< "1861.6431984064 16.0000000000"
				<< "1861.6620086572 0.0000000000"
				<< "1879.3452969428 0.0000000000"
				<< "1879.3642294765 27.0000000000"
				<< "1879.3831620103 0.0000000000"
				<< "2007.9375013022 0.0000000000"
				<< "2007.9570456000 34.0000000000"
				<< "2007.9765898979 3.0000000000"
				<< "2007.9962563479 0.0000000000"
				<< "2173.2127519882 0.0000000000"
				<< "2173.2330305731 52.0000000000"
				<< "2173.2533091582 3.0000000000";

			xyFormatStringListSize = xyFormatStringList.size();
			xyFormatString = xyFormatStringList.join("\n");
			MassSpectrum theorCombinedMassSpectrum;
			theorCombinedMassSpectrum.initialize(xyFormatString);
			REQUIRE(xyFormatStringListSize == theorCombinedMassSpectrum.size());

			xyFormatStringList.clear();

			// Finally actually test the combination function.

			MassSpectrum combinedMassSpectrum;

			combinedMassSpectrum.combine(massSpectra, mzStart, mzEnd);

			// Check that both the theor and the actual combined spectra have the same
			// number of mass peak instances.

			REQUIRE(theorCombinedMassSpectrum.size() == combinedMassSpectrum.size());

			// Now test that the obained mass combined mass spectrum is identical to the
			// one computed apart.

			int failedComparisons = 0;

			for(int iter = 0; iter < combinedMassSpectrum.size(); ++iter)
			{
				double mz = combinedMassSpectrum.at(iter)->mz();
				double i = combinedMassSpectrum.at(iter)->i();

				double theorMz = theorCombinedMassSpectrum.at(iter)->mz();
				double theorI = theorCombinedMassSpectrum.at(iter)->i();

				//cout << std::setprecision(15)
				//<< "(" << mz << "," << i << ") vs (" << theorMz << "," << theorI << ")\n";

				// It is important to use the almostEqual() because otherwise we get
				// spurious messages.
				if(!msXpS::almostEqual(mz,theorMz, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << mz << " vs theor " << theorMz << "\n";

					++failedComparisons;
				}

				if(!msXpS::almostEqual(i,theorI, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << i << " vs theor " << theorI << "\n";

					++failedComparisons;
				}
			}

			REQUIRE(failedComparisons == 0);
		}

		SECTION("Combination (not-binned case) of mzList/iList.", "[MassSpectrum]")
		{
			// We now have five mass spectra. We can now construct the mass spectrum
			// list. Only use the first four, as the fifth will be combined in the form
			// of mzList and iList, which allows use to recycle the combined spectrum
			// form an earlier test.

			MassSpectrumList massSpectra;
			massSpectra << massSpecIdx0 << massSpecIdx1 << massSpecIdx2 << massSpecIdx3;

			// Last spectrum to combine (massSpecIdx4) but in mzList/iList format:

			QList<double> mzList;
			mzList
				<< 1848.9427702443
				<< 1848.9615803966
				<< 1848.9803905489
				<< 1848.9990785576
				<< 2007.9375013022
				<< 2007.9570456000
				<< 2007.9765898979
				<< 2007.9962563479;

			QList<double> iList;
			iList
				<< 0.0
				<< 21.0
				<< 17.0
				<< 0.0
				<< 0.0
				<< 34.0
				<< 3.0
				<< 0.0;

			// The combined mass spectrum as computed apart:

			xyFormatStringList.clear();
			xyFormatStringList
				<< "1848.9427702443 0.0000000000"
				<< "1848.9615803966 21.0000000000"
				<< "1848.9803905489 17.0000000000"
				<< "1848.9990785576 0.0000000000"
				<< "1861.6054557607 0.0000000000"
				<< "1861.6243881557 4.0000000000"
				<< "1861.6431984064 16.0000000000"
				<< "1861.6620086572 0.0000000000"
				<< "1879.3452969428 0.0000000000"
				<< "1879.3642294765 27.0000000000"
				<< "1879.3831620103 0.0000000000"
				<< "2007.9375013022 0.0000000000"
				<< "2007.9570456000 34.0000000000"
				<< "2007.9765898979 3.0000000000"
				<< "2007.9962563479 0.0000000000"
				<< "2173.2127519882 0.0000000000"
				<< "2173.2330305731 52.0000000000"
				<< "2173.2533091582 3.0000000000"
				<< "2173.2738320637 0.0000000000"
				<< "2207.9724474677 0.0000000000"
				<< "2207.9932149877 3.0000000000"
				<< "2208.0137381840 32.0000000000"
				<< "2208.0342613806 0.0000000000";

			xyFormatStringListSize = xyFormatStringList.size();
			xyFormatString = xyFormatStringList.join("\n");
			MassSpectrum theorCombinedMassSpectrum;
			theorCombinedMassSpectrum.initialize(xyFormatString);
			REQUIRE(xyFormatStringListSize == theorCombinedMassSpectrum.size());

			xyFormatStringList.clear();

			// Finally actually test the combination function.

			MassSpectrum combinedMassSpectrum;
			combinedMassSpectrum.combine(massSpectra);
			combinedMassSpectrum.combine(mzList, iList);

			// Check that both the theor and the actual combined spectra have the same
			// number of mass peak instances.

			REQUIRE(theorCombinedMassSpectrum.size() == combinedMassSpectrum.size());

			// Now test that the obained mass combined mass spectrum is identical to the
			// one computed apart.

			int failedComparisons = 0;

			for(int iter = 0; iter < combinedMassSpectrum.size(); ++iter)
			{
				double mz = combinedMassSpectrum.at(iter)->mz();
				double i = combinedMassSpectrum.at(iter)->i();

				double theorMz = theorCombinedMassSpectrum.at(iter)->mz();
				double theorI = theorCombinedMassSpectrum.at(iter)->i();

				//cout << std::setprecision(15)
				//<< "(" << mz << "," << i << ") vs (" << theorMz << "," << theorI << ")\n";

				// It is important to use the almostEqual() because otherwise we get
				// spurious messages.
				if(!msXpS::almostEqual(mz,theorMz, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << mz << " vs theor " << theorMz << "\n";

					++failedComparisons;
				}

				if(!msXpS::almostEqual(i,theorI, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << i << " vs theor " << theorI << "\n";

					++failedComparisons;
				}
			}

			REQUIRE(failedComparisons == 0);
		}

		SECTION("Combination (not-binned case) of mzList/iList with mz range.", "[MassSpectrum]")
		{
			// We now have five mass spectra. We can now construct the mass spectrum
			// list. Only use the first four, as the fifth will be combined in the form
			// of mzList and iList, which allows use to recycle the combined spectrum
			// form an earlier test.

			MassSpectrumList massSpectra;
			massSpectra << massSpecIdx0 << massSpecIdx1 << massSpecIdx2 << massSpecIdx3;

			// Last spectrum to combine (massSpecIdx4) but in mzList/iList format:

			QList<double> mzList;
			mzList
				<< 1848.9427702443
				<< 1848.9615803966
				<< 1848.9803905489
				<< 1848.9990785576
				<< 2007.9375013022
				<< 2007.9570456000
				<< 2007.9765898979
				<< 2007.9962563479;

			QList<double> iList;
			iList
				<< 0.0
				<< 21.0
				<< 17.0
				<< 0.0
				<< 0.0
				<< 34.0
				<< 3.0
				<< 0.0;

			// The combined mass spectrum as computed apart:

			xyFormatStringList.clear();
			xyFormatStringList
				<< "1861.6243881557 4.0000000000"
				<< "1861.6431984064 16.0000000000"
				<< "1861.6620086572 0.0000000000"
				<< "1879.3452969428 0.0000000000"
				<< "1879.3642294765 27.0000000000"
				<< "1879.3831620103 0.0000000000"
				<< "2007.9375013022 0.0000000000"
				<< "2007.9570456000 34.0000000000"
				<< "2007.9765898979 3.0000000000"
				<< "2007.9962563479 0.0000000000"
				<< "2173.2127519882 0.0000000000"
				<< "2173.2330305731 52.0000000000"
				<< "2173.2533091582 3.0000000000";

			xyFormatStringListSize = xyFormatStringList.size();
			xyFormatString = xyFormatStringList.join("\n");
			MassSpectrum theorCombinedMassSpectrum;
			theorCombinedMassSpectrum.initialize(xyFormatString);
			REQUIRE(xyFormatStringListSize == theorCombinedMassSpectrum.size());

			xyFormatStringList.clear();

			// Skip 5 spectra on the top (massSpecIdx0).
			// A mass peak is in the range if mz >= mzStart.
			double mzStart = 1861.61;

			// Skip 5 spectra on the bottom
			// A mass peak is in the range if mz <= mzEnd.
			double mzEnd = 2173.26;

			// Finally actually test the combination function.

			MassSpectrum combinedMassSpectrum;
			combinedMassSpectrum.combine(massSpectra, mzStart, mzEnd);
			combinedMassSpectrum.combine(mzList, iList, mzStart, mzEnd);

			// Check that both the theor and the actual combined spectra have the same
			// number of mass peak instances.

			REQUIRE(theorCombinedMassSpectrum.size() == combinedMassSpectrum.size());

			// Now test that the obained mass combined mass spectrum is identical to the
			// one computed apart.

			int failedComparisons = 0;

			for(int iter = 0; iter < combinedMassSpectrum.size(); ++iter)
			{
				double mz = combinedMassSpectrum.at(iter)->mz();
				double i = combinedMassSpectrum.at(iter)->i();

				double theorMz = theorCombinedMassSpectrum.at(iter)->mz();
				double theorI = theorCombinedMassSpectrum.at(iter)->i();

				//cout << std::setprecision(15)
				//<< "(" << mz << "," << i << ") vs (" << theorMz << "," << theorI << ")\n";

				// It is important to use the almostEqual() because otherwise we get
				// spurious messages.
				if(!msXpS::almostEqual(mz,theorMz, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << mz << " vs theor " << theorMz << "\n";

					++failedComparisons;
				}

				if(!msXpS::almostEqual(i,theorI, 15))
				{
					cout << std::setprecision(15) << "failed comparison:"
						<< "test (" << i << " vs theor " << theorI << "\n";

					++failedComparisons;
				}
			}

			REQUIRE(failedComparisons == 0);
		}

		delete massSpecIdx0;
		delete massSpecIdx1;
		delete massSpecIdx2;
		delete massSpecIdx3;
		delete massSpecIdx4; 

		// At this point we have tested the un-binned combination into an empty mass
		// spectrum of a mass spectrum list containing five mass spectra.
	}


	TEST_CASE("Various utility functions about mass spectra.", "[MassSpectrum]")
	{
		QStringList xyFormatStringList;
		xyFormatStringList
			<< "694.5165997465 1500.0000000000"
			<< "694.5333557712 1332.0000000000"
			<< "694.5501119980 1410.0000000000"
			<< "694.5668684269 1363.0000000000"
			<< "694.5836250580 1424.0000000000"
			<< "694.6003818912 1503.0000000000"
			<< "694.6171389265 1341.0000000000"
			<< "694.6338961639 1089.0000000000"
			<< "694.6506536035 983.0000000000"
			<< "694.6674112451 999.0000000000"
			<< "694.6841690890 1043.0000000000"
			<< "694.7009271349 1015.0000000000"
			<< "694.7176853830 1057.0000000000"
			<< "694.7344438331 1140.0000000000"
			<< "694.7512024855 1284.0000000000"
			<< "694.7679613399 1479.0000000000"
			<< "694.7847203965 1661.0000000000"
			<< "694.8014796552 1703.0000000000"
			<< "694.8182391160 1541.0000000000"
			<< "694.8349987789 879.0000000000" ;

		REQUIRE(xyFormatStringList.size() == 20);
		QString xyFormatString = xyFormatStringList.join("\n");
		MassSpectrum massSpectrum;
		massSpectrum.initialize(xyFormatString);
		REQUIRE(xyFormatStringList.size() == massSpectrum.size());

		SECTION("Tell if a mass spectrum contains a given m/z value.", "[MassSpectrum]")
		{
			int index = 10;
			double mz = massSpectrum.at(index)->mz();

			// 20170926 add one more decimal to 0.0000000000001 and it fails.
			REQUIRE(massSpectrum.contains(mz + 0.0000000000001) < 0);
			REQUIRE(massSpectrum.contains(mz) >= 0);
			REQUIRE(massSpectrum.contains(mz) == 10);
		}

		SECTION("Provide the mass spectral m/z data as a list of double values.", "[MassSpectrum]")
		{
			QList<double> mzList = massSpectrum.mzList();

			int failedComparisons = 0;

			for(int iter = 0; iter < massSpectrum.size(); ++iter)
			{
				double iterMz = massSpectrum.at(iter)->mz();

				double listMz = mzList.at(iter);

				if(iterMz != listMz)
					++failedComparisons;
			}	

			REQUIRE(failedComparisons == 0);
		}

		SECTION("Provide the mass spectral i data as a list of double values.", "[MassSpectrum]")
		{
			QList<double> iList = massSpectrum.iList();

			int failedComparisons = 0;

			for(int iter = 0; iter < massSpectrum.size(); ++iter)
			{
				double iterI = massSpectrum.at(iter)->i();

				double listI = iList.at(iter);

				if(iterI != listI)
					++failedComparisons;
			}	

			REQUIRE(failedComparisons == 0);
		}

		SECTION("Provide the mass spectral m/z and i data as two lists of double values.", "[MassSpectrum]")
		{
			QList<double> mzList;
			QList<double> iList;

			massSpectrum.toLists(&mzList, &iList);

			int failedComparisons = 0;

			for(int iter = 0; iter < massSpectrum.size(); ++iter)
			{
				double iterMz = massSpectrum.at(iter)->mz();
				double listMz = mzList.at(iter);

				double iterI = massSpectrum.at(iter)->i();
				double listI = iList.at(iter);

				if(iterMz != listMz)
					++failedComparisons;
				if(iterI != listI)
					++failedComparisons;
			}	

			REQUIRE(failedComparisons == 0);
		}

		SECTION("Provide the mass spectral m/z and i data as two vectors of double values.", "[MassSpectrum]")
		{
			QList<double> mzList;
			QList<double> iList;

			// We have tested that this works correctly.
			massSpectrum.toLists(&mzList, &iList);

			QVector<double> mzVector;
			QVector<double> iVector;

			massSpectrum.toVectors(&mzVector, &iVector);

			int failedComparisons = 0;

			for(int iter = 0; iter < massSpectrum.size(); ++iter)
			{
				double iterMz = massSpectrum.at(iter)->mz();
				double vectorMz = mzVector.at(iter);

				double iterI = massSpectrum.at(iter)->i();
				double vectorI = iVector.at(iter);

				if(iterMz != vectorMz)
					++failedComparisons;
				if(iterI != vectorI)
					++failedComparisons;
			}	

			REQUIRE(failedComparisons == 0);
		}

		SECTION("Provide the mass spectral m/z and i data as a <m/z,i> map of double values.", "[MassSpectrum]")
		{
			QList<double> mzList;
			QList<double> iList;

			// We have tested that this works correctly.
			massSpectrum.toLists(&mzList, &iList);

			QMap<double, double> map;
			massSpectrum.toMap(&map);

			int failedComparisons = 0;

			QList<double> mapMzList = map.keys();

			REQUIRE(mapMzList.size() == mzList.size());

			for(int iter = 0; iter < massSpectrum.size(); ++iter)
			{
				double iterMz = massSpectrum.at(iter)->mz();
				double mapMz = mapMzList.at(iter);

				double iterI = massSpectrum.at(iter)->i();
				double mapI = map.value(mapMz);

				if(iterMz != mapMz)
					++failedComparisons;
				if(iterI != mapI)
					++failedComparisons;
			}	

			REQUIRE(failedComparisons == 0);
		}

		SECTION("Create a textual representation of a region of a mass "
				"spectrum (without mz range).", "[MassSpectrum]")
		{
			QString *text = massSpectrum.asText();
			REQUIRE(text != Q_NULLPTR);

			MassSpectrum reconstructedMassSpectrum;
			reconstructedMassSpectrum.initialize(*text);

			delete text;

			REQUIRE(massSpectrum.size() == reconstructedMassSpectrum.size());

			// Now perform a deep comparison of the spectra.
			int failedComparisons = 0;

			for(int iter = 0; iter < massSpectrum.size(); ++iter)
			{
				double iterMz = massSpectrum.at(iter)->mz();
				double reconstructMz = reconstructedMassSpectrum.at(iter)->mz();

				double iterI = massSpectrum.at(iter)->i();
				double reconstructI = reconstructedMassSpectrum.at(iter)->i();

				if(iterMz != reconstructMz)
					++failedComparisons;
				if(iterI != reconstructI)
					++failedComparisons;
			}	

			REQUIRE(failedComparisons == 0);
		}

		SECTION("Create a textual representation of a region of a mass "
				"spectrum (with mz range).", "[MassSpectrum]")
		{

			// Now do the same with a mz range. We know that massSpectrum contains this:
			//
			//<< "694.5165997465 1500.0000000000"
			//<< "694.5333557712 1332.0000000000"
			//<< "694.5501119980 1410.0000000000"
			//<< "694.5668684269 1363.0000000000"
			//<< "694.5836250580 1424.0000000000"
			//<< "694.6003818912 1503.0000000000"
			//<< "694.6171389265 1341.0000000000"
			//<< "694.6338961639 1089.0000000000"
			//<< "694.6506536035 983.0000000000"
			//<< "694.6674112451 999.0000000000"
			//<< "694.6841690890 1043.0000000000"
			//<< "694.7009271349 1015.0000000000"
			//<< "694.7176853830 1057.0000000000"
			//<< "694.7344438331 1140.0000000000"
			//<< "694.7512024855 1284.0000000000"
			//<< "694.7679613399 1479.0000000000"
			//<< "694.7847203965 1661.0000000000"
			//<< "694.8014796552 1703.0000000000"
			//<< "694.8182391160 1541.0000000000"
			//<< "694.8349987789 879.0000000000" ;

			// Skip the 5 top-most and bottom-most peaks.
			QString *text = massSpectrum.asText(5, 14);
			REQUIRE(text != Q_NULLPTR);

			MassSpectrum reconstructedMassSpectrum;
			reconstructedMassSpectrum.initialize(*text);

			delete text;

			// The reconstructed mass spectrum has 10 peaks less than the original
			// one.
			REQUIRE(massSpectrum.size() == reconstructedMassSpectrum.size() + 10);

			// Now perform a deep comparison of the spectra. Note the dephasing during
			// the iteration in both spectra.
			int failedComparisons = 0;

			for(int iter = 0; iter < reconstructedMassSpectrum.size(); ++iter)
			{
				double iterMz = massSpectrum.at(iter + 5)->mz();
				double reconstructMz = reconstructedMassSpectrum.at(iter)->mz();

				double iterI = massSpectrum.at(iter + 5)->i();
				double reconstructI = reconstructedMassSpectrum.at(iter)->i();

				if(iterMz != reconstructMz)
					++failedComparisons;
				if(iterI != reconstructI)
					++failedComparisons;
			}	

			REQUIRE(failedComparisons == 0);
		}
	}
	// End of
	// TEST_CASE("Various utility functions about mass spectra.", "[MassSpectrum]")


	TEST_CASE("Create mass spectrum from base64-encoded data in byte arrays.", "[MassSpectrum]")
	{
		// Prepare the byte arrays from reading data from specific test files.

		QFile mz64File(QString("%1/base64/protein-mobility-waters-synapt-hdms.db-index-47-mzlist.base64")
				.arg(TEST_DATA_DIR));
		REQUIRE(mz64File.open(QFile::ReadOnly));
		QTextStream mz64Stream(&mz64File);
		QByteArray mz64ByteArray = mz64Stream.readAll().toLatin1();
		QByteArray mzDec64ByteArray = QByteArray::fromBase64(mz64ByteArray);

		QFile i64File(QString("%1/base64/protein-mobility-waters-synapt-hdms.db-index-47-ilist.base64")
				.arg(TEST_DATA_DIR));
		REQUIRE(i64File.open(QFile::ReadOnly));
		QTextStream i64Stream(&i64File);
		QByteArray i64ByteArray = i64Stream.readAll().toLatin1();
		QByteArray iDec64ByteArray = QByteArray::fromBase64(i64ByteArray);

		// We know that the files correspond to zlib-compressed data, so the size of
		// the encoded byte arrays can be different (and must in this specific
		// case).
		REQUIRE(mzDec64ByteArray.size() != iDec64ByteArray.size());

		// As per Qt documentation is decompressing zlib-compressed data from
		// outside of Qt, then we have to prepend 4 bytes to the arrays before
		// decompression.
		QByteArray header("aaaa");
		mzDec64ByteArray.prepend(header);
		iDec64ByteArray.prepend(header);

		QByteArray finalMzArray = qUncompress(mzDec64ByteArray);
		QByteArray finalIArray = qUncompress(iDec64ByteArray);

		// Now , yes, after decompression, the number of mz items needs to be
		// identical to the number of i items.
		REQUIRE(finalMzArray.size() == finalIArray.size());

		MassSpectrum *massSpectrum = MassSpectrum::createFromBase64EncByteArrays(&mz64ByteArray, &i64ByteArray,
				msXpS::DataCompression::DATA_COMPRESSION_ZLIB);
		REQUIRE(massSpectrum != Q_NULLPTR);

		// This one contains both mz and i lists in the xy format.
		QFile xyFile(QString("%1/xy/protein-mobility-waters-synapt-hdms.db-spectrum-index-47.xy")
				.arg(TEST_DATA_DIR));
		REQUIRE(xyFile.open(QFile::ReadOnly));
		QTextStream xyStream(&xyFile);
		QString xyText = xyStream.readAll();

		MassSpectrum testSpectrum;
		testSpectrum.initialize(xyText);

		REQUIRE(massSpectrum->size() == testSpectrum.size());

		int failedComparisons = 0;

		for(int iter = 0; iter < testSpectrum.size(); ++iter)
		{
			double mz = massSpectrum->at(iter)->mz();
			double testMz = testSpectrum.at(iter)->mz();

			double i = massSpectrum->at(iter)->i();
			double testI = testSpectrum.at(iter)->i();

			if(abs(mz-testMz) > 0.0000000001)
				//if(!msXpS::almostEqual(mz, testMz, 10))
			{
				cout << std::setprecision(10) << mz << " vs " << testMz << "\n"
					"\t\t difference:" << mz - testMz << "\n";

				++failedComparisons;
			}

			//if(!msXpS::almostEqual(i, testI, 10))
			//++failedComparisons;
		}

		REQUIRE(failedComparisons == 0);

		delete massSpectrum;
	}
	// End of 
	// TEST_CASE("Create mass spectrum from base64-encoded data in byte arrays.", "[MassSpectrum]")


	TEST_CASE("Create mass spectrum from base64-encoded data in "
			"byte arrays (with mz range).", "[MassSpectrum]")
	{
		// Prepare the byte arrays from reading data from specific test files.

		QFile mz64File(QString("%1/base64/protein-mobility-waters-synapt-hdms.db-index-47-mz-2000-2400-mzlist.base64")
				.arg(TEST_DATA_DIR));
		REQUIRE(mz64File.open(QFile::ReadOnly));
		QTextStream mz64Stream(&mz64File);
		QByteArray mz64ByteArray = mz64Stream.readAll().toLatin1();
		QByteArray mzDec64ByteArray = QByteArray::fromBase64(mz64ByteArray);

		QFile i64File(QString("%1/base64/protein-mobility-waters-synapt-hdms.db-index-47-mz-2000-2400-ilist.base64")
				.arg(TEST_DATA_DIR));
		REQUIRE(i64File.open(QFile::ReadOnly));
		QTextStream i64Stream(&i64File);
		QByteArray i64ByteArray = i64Stream.readAll().toLatin1();
		QByteArray iDec64ByteArray = QByteArray::fromBase64(i64ByteArray);

		// We know that the files correspond to zlib-compressed data, so the size of
		// the encoded byte arrays can be different (and must in this specific
		// case).
		REQUIRE(mzDec64ByteArray.size() != iDec64ByteArray.size());

		// As per Qt documentation is decompressing zlib-compressed data from
		// outside of Qt, then we have to prepend 4 bytes to the arrays before
		// decompression.
		QByteArray header("aaaa");
		mzDec64ByteArray.prepend(header);
		iDec64ByteArray.prepend(header);

		QByteArray finalMzArray = qUncompress(mzDec64ByteArray);
		QByteArray finalIArray = qUncompress(iDec64ByteArray);

		// Now , yes, after decompression, the number of mz items needs to be
		// identical to the number of i items.
		REQUIRE(finalMzArray.size() == finalIArray.size());

		MassSpectrum *massSpectrum = MassSpectrum::createFromBase64EncByteArrays(&mz64ByteArray, &i64ByteArray,
				msXpS::DataCompression::DATA_COMPRESSION_ZLIB, 2000, 2400);
		REQUIRE(massSpectrum != Q_NULLPTR);

		// This one contains both mz and i lists in the xy format.
		QFile xyFile(QString("%1/xy/protein-mobility-waters-synapt-hdms.db-spectrum-index-47-mz-2000-2400.xy")
				.arg(TEST_DATA_DIR));
		REQUIRE(xyFile.open(QFile::ReadOnly));
		QTextStream xyStream(&xyFile);
		QString xyText = xyStream.readAll();

		MassSpectrum testSpectrum;
		testSpectrum.initialize(xyText);

		REQUIRE(massSpectrum->size() == testSpectrum.size());

		int failedComparisons = 0;

		for(int iter = 0; iter < testSpectrum.size(); ++iter)
		{
			double mz = massSpectrum->at(iter)->mz();
			double testMz = testSpectrum.at(iter)->mz();

			double i = massSpectrum->at(iter)->i();
			double testI = testSpectrum.at(iter)->i();

			if(abs(mz-testMz) > 0.0000000001)
				//if(!msXpS::almostEqual(mz, testMz, 10))
			{
				cout << std::setprecision(10) << mz << " vs " << testMz << "\n"
					"\t\t difference:" << mz - testMz << "\n";

				++failedComparisons;
			}

			//if(!msXpS::almostEqual(i, testI, 10))
			//++failedComparisons;
		}

		REQUIRE(failedComparisons == 0);

		delete massSpectrum;
	}





	//TEST_CASE("", "[MassSpectrum]")
	//{

	//}

} // namespace msXpSlibmass
